# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for qcookiecutter widget.
"""

# Standard library imports
import json
import os
import shutil
import tempfile

# Third party imports
import pytest

# Local imports
from spyder.plugins.projects.utils.cookie import (
    generate_cookiecutter_project, load_cookiecutter_project)


def test_load_cookiecutter_project_config():
    settings = {
        "opt_1": "value",
        "opt_2": "{{ cookiecutter.opt_1 }}",
    }
    temp_path = tempfile.mkdtemp(suffix='-some-cookiecutter')
    temp_cookie_path = os.path.join(temp_path, 'cookiecutter.json')

    with open(temp_cookie_path, 'w') as fh:
        fh.write(json.dumps(settings, sort_keys=True))

    sets, pre_gen_code = load_cookiecutter_project(temp_path)
    assert settings == sets
    assert pre_gen_code is None

    shutil.rmtree(temp_path)


def test_load_cookiecutter_project_hooks():
    settings = {
        "opt_1": "value",
        "opt_2": "{{ cookiecutter.opt_1 }}",
    }
    pre_gen_code = "import sys\n\nprint('test!')\nsys.exit(1)\n"
    temp_path = tempfile.mkdtemp(suffix='-some-cookiecutter')
    temp_cookie_path = os.path.join(temp_path, 'cookiecutter.json')
    temp_hooks_path = os.path.join(temp_path, 'hooks')
    temp_hooks_pre_path = os.path.join(temp_hooks_path, 'pre_gen_project.py')
    os.makedirs(temp_hooks_path)

    with open(temp_cookie_path, 'w') as fh:
        fh.write(json.dumps(settings, sort_keys=True))

    with open(temp_hooks_pre_path, 'w') as fh:
        fh.write(pre_gen_code)

    sets, pre_gen_code = load_cookiecutter_project(temp_path)
    assert settings == sets
    assert pre_gen_code == pre_gen_code

    shutil.rmtree(temp_path)


def test_generate_cookiecutter_project_defaults():
    settings = {
        "repo_name": "value",
    }
    temp_path = tempfile.mkdtemp(suffix='-some-cookiecutter')
    temp_path_created = tempfile.mkdtemp(suffix='-created-project')
    temp_cookie_path = os.path.join(temp_path, 'cookiecutter.json')
    temp_project_path = os.path.join(temp_path, '{{cookiecutter.repo_name}}')
    os.makedirs(temp_project_path)

    with open(temp_cookie_path, 'w') as fh:
        fh.write(json.dumps(settings, sort_keys=True))

    status, result = generate_cookiecutter_project(
        temp_path,
        temp_path_created,
    )
    assert "value" in result
    assert status is True
    shutil.rmtree(temp_path)


def test_generate_cookiecutter_project_extra_content():
    settings = {
        "repo_name": "value",
    }
    temp_path = tempfile.mkdtemp(suffix='-some-cookiecutter')
    temp_path_created = tempfile.mkdtemp(suffix='-created-project')
    temp_cookie_path = os.path.join(temp_path, 'cookiecutter.json')
    temp_project_path = os.path.join(temp_path, '{{cookiecutter.repo_name}}')
    os.makedirs(temp_project_path)

    with open(temp_cookie_path, 'w') as fh:
        fh.write(json.dumps(settings, sort_keys=True))

    status, result = generate_cookiecutter_project(
        temp_path,
        temp_path_created,
        {"repo_name": "boom"},
    )
    assert "boom" in result
    assert status is True
    shutil.rmtree(temp_path)


def test_generate_cookiecutter_project_exception():
    settings = {
        "repo_name": "value",
    }
    temp_path = tempfile.mkdtemp(suffix='-some-invalid-cookiecutter')
    temp_path_created = tempfile.mkdtemp(suffix='-created-project')
    temp_cookie_path = os.path.join(temp_path, 'cookiecutter.json')
    temp_project_path = os.path.join(
        temp_path,
        '{{cookiecutter.not_foun_variable}}',
    )
    os.makedirs(temp_project_path)

    with open(temp_cookie_path, 'w') as fh:
        fh.write(json.dumps(settings, sort_keys=True))

    status, __ = generate_cookiecutter_project(
        temp_path,
        temp_path_created,
    )
    assert status is False


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for environ.py
"""

# Standard library imports
import os

# Test library imports
import pytest

# Third party imports
from qtpy.QtCore import QTimer

# Local imports
from spyder.utils.environ import (get_user_environment_variables,
                                  UserEnvDialog, amend_user_shell_init)
from spyder.utils.test import close_message_box
from spyder.app.tests.conftest import restore_user_env


@pytest.fixture
def environ_dialog(qtbot):
    "Setup the Environment variables Dialog."
    QTimer.singleShot(1000, lambda: close_message_box(qtbot))
    dialog = UserEnvDialog()
    qtbot.addWidget(dialog)

    return dialog


def test_get_user_environment_variables():
    """Test get_user_environment_variables function"""
    # All platforms should have a path environment variable, but
    # Windows may have mixed case.
    keys = {k.lower() for k in get_user_environment_variables()}
    assert "path" in keys or "shlvl" in keys


@pytest.mark.skipif(os.name == "nt", reason="Does not apply to Windows")
def test_get_user_env_newline(restore_user_env):
    # Test variable value with newline characters.
    # Regression test for spyder-ide#20097
    text = "myfunc() {  echo hello;\n echo world\n}\nexport -f myfunc"
    amend_user_shell_init(text)
    user_env = get_user_environment_variables()

    assert user_env['BASH_FUNC_myfunc%%'] in text


def test_environ(environ_dialog, qtbot):
    """Test the environment variables dialog."""
    environ_dialog.show()
    assert environ_dialog

    # All platforms should have a path environment variable, but
    # Windows may have mixed case.
    keys = {k.lower() for k in environ_dialog.get_value()}
    assert "path" in keys or "shlvl" in keys


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for misc.py
"""

# Standard library imports
import os

# Test library imports
import pytest

# Local imports
from spyder.utils.misc import get_common_path


def test_get_common_path():
    """Test getting the common path."""
    if os.name == 'nt':
        assert get_common_path(['D:\\Python\\spyder-v21\\spyder\\widgets',
                                'D:\\Python\\spyder\\spyder\\utils',
                                'D:\\Python\\spyder\\spyder\\widgets',
                                'D:\\Python\\spyder-v21\\spyder\\utils',
                                ]) == 'D:\\Python'
    else:
        assert get_common_path(['/Python/spyder-v21/spyder.widgets',
                                '/Python/spyder/spyder.utils',
                                '/Python/spyder/spyder.widgets',
                                '/Python/spyder-v21/spyder.utils',
                                ]) == '/Python'


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Dialog window notifying user of autosave-related errors."""

# Standard library imports
import logging

# Third party imports
from qtpy.QtWidgets import (QCheckBox, QDialog, QDialogButtonBox, QLabel,
                            QVBoxLayout)

# Local imports
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.config.base import _


logger = logging.getLogger(__name__)


class AutosaveErrorDialog(QDialog):
    """
    Dialog window notifying user of autosave-related errors.

    The window also includes a check box which allows the user to hide any
    future autosave-related errors.

    Class attribute:
        show_errors (bool): whether to show errors or not
    """

    show_errors = True

    def __init__(self, action, error):
        """
        Constructor.

        Args:
            action (str): what Spyder was trying to do when error occurred
            error (Exception): the error that occurred
        """
        logger.error(action, exc_info=error)

        QDialog.__init__(self)
        self.setWindowTitle(_('Autosave error'))
        self.setModal(True)

        layout = QVBoxLayout()

        header = _('Error message:')
        txt = '<br>{}<br><br>{}<br>{!s}'.format(action, header, error)
        layout.addWidget(QLabel(txt))
        layout.addSpacing(15)

        txt = _("Hide all future autosave-related errors during this session")
        self.dismiss_box = QCheckBox(txt)
        layout.addWidget(self.dismiss_box)
        layout.addSpacing(15)

        button_box = SpyderDialogButtonBox(QDialogButtonBox.Ok)
        button_box.accepted.connect(self.accept)
        layout.addWidget(button_box)

        self.setLayout(layout)

    def exec_if_enabled(self):
        """
        Execute dialog box unless disabled by the user.

        The dialog box is disabled once the user clicks the 'Hide all future
        errors' check box on one dialog box.
        """
        if AutosaveErrorDialog.show_errors:
            return self.exec_()

    def accept(self):
        """
        Update `show_errors` and hide dialog box.

        Overrides method of `QDialogBox`.
        """
        AutosaveErrorDialog.show_errors = not self.dismiss_box.isChecked()
        return QDialog.accept(self)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Switcher Plugin.
"""

# Standard library imports
import sys

# Third-party imports
from qtpy.QtCore import Signal

# Local imports
from spyder.api.translations import _
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.plugin_registration.decorators import (on_plugin_available,
                                                       on_plugin_teardown)
from spyder.plugins.switcher.api import SwitcherActions
from spyder.plugins.switcher.container import SwitcherContainer
from spyder.plugins.mainmenu.api import ApplicationMenus, FileMenuSections


class Switcher(SpyderPluginV2):
    """
    Switcher plugin.
    """

    NAME = "switcher"
    OPTIONAL = [Plugins.MainMenu]
    CONTAINER_CLASS = SwitcherContainer
    CONF_SECTION = NAME
    CONF_FILE = False

    # --- Signals
    sig_rejected = Signal()
    """
    This signal is emitted when the plugin is dismissed.
    """

    sig_item_changed = Signal(object)
    """
    This signal is emitted when the current item changes.
    """

    sig_item_selected = Signal(object, str, str)
    """
    This signal is emitted when an item is selected from the switcher plugin
    list.

    Parameters
    ----------
    item: object
        The current selected item from the switcher list (QStandardItem).
    mode: str
        The current selected mode (open files "", symbol "@" or line ":").
    search_text: str
        Cleaned search/filter text.
    """

    sig_mode_selected = Signal(str)
    """
    This signal is emitted when a mode is selected.

    Parameters
    ----------
    mode: str
        The selected mode (open files "", symbol "@" or line ":").
    """

    sig_search_text_available = Signal(str)
    """
    This signal is emitted when the user stops typing the search/filter text.

    Parameters
    ----------
    search_text: str
        The current search/filter text.
    """

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Switcher")

    @staticmethod
    def get_description():
        return _("Quickly switch between files and other items.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('switcher')

    def on_initialize(self):
        container = self.get_container()
        self._switcher = container.switcher

        self._switcher.sig_rejected.connect(self.sig_rejected)
        self._switcher.sig_item_changed.connect(self.sig_item_changed)
        self._switcher.sig_item_selected.connect(self.sig_item_selected)
        self._switcher.sig_mode_selected.connect(self.sig_mode_selected)
        self._switcher.sig_search_text_available.connect(
            self.sig_search_text_available
        )

    def on_close(self, cancellable=True):
        """Close switcher widget."""
        self._switcher.close()

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for switcher_action in [
                SwitcherActions.FileSwitcherAction,
                SwitcherActions.SymbolFinderAction]:
            action = self.get_action(switcher_action)
            if sys.platform == 'darwin':
                before_section = FileMenuSections.Navigation
            else:
                before_section = FileMenuSections.Restart
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Switcher,
                before_section=before_section
            )

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        for switcher_action in [
                SwitcherActions.FileSwitcherAction,
                SwitcherActions.SymbolFinderAction]:
            action = self.get_action(switcher_action)
            mainmenu.remove_item_from_application_menu(
                action,
                menu_id=ApplicationMenus.File
            )

    # ---- Public API
    # -------------------------------------------------------------------------
    # --- Switcher methods
    def set_placeholder_text(self, text):
        """Set the text appearing on the empty line edit."""
        self._switcher.set_placeholder_text(text)

    def setup(self):
        """Setup list widget content based on filtering."""
        self._switcher.setup()

    def open_switcher(self, symbol=False):
        """Open switcher dialog."""
        self.get_container().open_switcher(symbol)

    def open_symbolfinder(self):
        """Open symbol list management dialog."""
        self.get_container().open_symbolfinder()

    # --- QDialog methods
    def show(self):
        """Show switcher."""
        self._switcher.show()

    def hide(self):
        """Hide switcher."""
        self._switcher.hide()

    def set_visible(self, visible):
        """Show or hide switcher."""
        self._switcher.setVisible(visible)

    def is_visible(self):
        """Return if the switcher is visible."""
        return self._switcher.isVisible()

    # --- Item methods
    def current_item(self):
        """Return the current selected item in the list widget."""
        return self._switcher.current_item()

    def add_item(self, icon=None, title=None, description=None, shortcut=None,
                 section=None, data=None, tool_tip=None, action_item=False,
                 last_item=True, score=-1, use_score=True):
        """Add a switcher list item."""
        self._switcher.add_item(icon, title, description, shortcut,
                                section, data, tool_tip, action_item,
                                last_item, score, use_score)

    def set_current_row(self, row):
        """Set the current selected row in the switcher."""
        self._switcher.set_current_row(row)

    def add_separator(self):
        """Add a separator item."""
        self._switcher.add_separator()

    def clear(self):
        """Remove all items from the list and clear the search text."""
        self._switcher.clear()

    def count(self):
        """Get the item count in the list widget."""
        return self._switcher.count()

    def remove_section(self, section):
        """Remove all items in a section of the switcher."""
        self._switcher.remove_section(section)

    # --- Mode methods
    def add_mode(self, token, description):
        """Add mode by token key and description."""
        self._switcher.add_mode(token, description)

    def get_mode(self):
        """Get the current mode the switcher is in."""
        return self._switcher.get_mode()

    def remove_mode(self, token):
        """Remove mode by token key."""
        self._switcher.remove_mode(token)

    def clear_modes(self):
        """Delete all modes spreviously defined."""
        self._switcher.clear_modes()

    # --- Lineedit methods
    def set_search_text(self, string):
        """Set the content of the search text."""
        self._switcher.set_search_text(string)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Remote Client Plugin.
"""

# Standard library imports
import logging
import contextlib

# Third-party imports
from qtpy.QtCore import Signal, Slot

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.plugin_registration.decorators import (
    on_plugin_available,
    on_plugin_teardown,
)
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.translations import _
from spyder.plugins.ipythonconsole.api import IPythonConsoleWidgetActions
from spyder.plugins.mainmenu.api import (
    ApplicationMenus,
    ConsolesMenuSections,
    ToolsMenuSections,
)
from spyder.plugins.remoteclient.api import (
    RemoteClientActions,
    RemoteClientMenus,
)
from spyder.plugins.remoteclient.api.client import SpyderRemoteClient
from spyder.plugins.remoteclient.api.protocol import (
    SSHClientOptions,
    ConnectionStatus,
)
from spyder.plugins.remoteclient.widgets.container import RemoteClientContainer

_logger = logging.getLogger(__name__)


class RemoteClient(SpyderPluginV2):
    """
    Remote client plugin.
    """

    NAME = "remoteclient"
    OPTIONAL = [Plugins.IPythonConsole, Plugins.MainMenu]
    CONF_SECTION = NAME
    CONTAINER_CLASS = RemoteClientContainer
    CONF_FILE = False

    CONF_SECTION_SERVERS = "servers"

    # ---- Signals
    sig_server_stopped = Signal(str)
    sig_server_renamed = Signal(str)
    sig_client_message_logged = Signal(dict)

    sig_connection_established = Signal(str)
    sig_connection_lost = Signal(str)
    sig_connection_status_changed = Signal(dict)

    sig_version_mismatch = Signal(str, str)

    _sig_kernel_started = Signal(object, dict)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._remote_clients: dict[str, SpyderRemoteClient] = {}

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Remote client")

    @staticmethod
    def get_description():
        return _("Connect to remote machines to run code on them.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon("remote_server")

    def on_initialize(self):
        self._reset_status()
        self._is_consoles_menu_added = False
        container = self.get_container()

        # Container signals
        container.sig_start_server_requested.connect(self.start_remote_server)
        container.sig_stop_server_requested.connect(self.stop_remote_server)
        container.sig_stop_server_requested.connect(self.sig_server_stopped)
        container.sig_server_renamed.connect(self.sig_server_renamed)
        container.sig_create_ipyclient_requested.connect(
            self.create_ipyclient_for_server
        )
        container.sig_shutdown_kernel_requested.connect(self._shutdown_kernel)
        container.sig_interrupt_kernel_requested.connect(
            self._interrupt_kernel
        )

        # Plugin signals
        self.sig_connection_status_changed.connect(
            container.sig_connection_status_changed
        )
        self.sig_client_message_logged.connect(
            container.sig_client_message_logged
        )
        self.sig_version_mismatch.connect(
            container.on_server_version_mismatch
        )
        self._sig_kernel_started.connect(container.on_kernel_started)

    def on_first_registration(self):
        pass

    def on_close(self, cancellable=True):
        """Stops remote server and close any opened connection."""
        for client in self._remote_clients.values():
            AsyncDispatcher(
                client.close, loop="asyncssh", early_return=False
            )()

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_mainmenu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        action = self.get_action(RemoteClientActions.ManageConnections)
        mainmenu.add_item_to_application_menu(
            action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.External,
            before_section=ToolsMenuSections.Extras,
        )

        if (
            self.is_plugin_available(Plugins.IPythonConsole)
            and not self._is_consoles_menu_added
        ):
            self._add_remote_consoles_menu()

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_mainmenu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        mainmenu.remove_item_from_application_menu(
            RemoteClientActions.ManageConnections,
            menu_id=ApplicationMenus.Tools,
        )

        if self._is_consoles_menu_added:
            mainmenu.remove_item_from_application_menu(
                RemoteClientMenus.RemoteConsoles,
                menu_id=ApplicationMenus.Consoles,
            )

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipython_console_available(self):
        if (
            self.is_plugin_available(Plugins.MainMenu)
            and not self._is_consoles_menu_added
        ):
            self._add_remote_consoles_menu()

    # ---- Public API
    # -------------------------------------------------------------------------
    # --- Remote Server Methods
    def get_remote_server(self, config_id):
        """Get remote server."""
        if config_id in self._remote_clients:
            return self._remote_clients[config_id]

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _install_remote_server(self, config_id):
        """Install remote server."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.connect_and_install_remote_server()

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def start_remote_server(self, config_id):
        """Start remote server."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]
        await client.connect_and_ensure_server()

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def stop_remote_server(self, config_id):
        """Stop remote server."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.close()

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def ensure_remote_server(self, config_id):
        """Ensure remote server is running and installed."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.connect_and_ensure_server()

    def restart_remote_server(self, config_id):
        """Restart remote server."""
        self.stop_remote_server(config_id)
        self.start_remote_server(config_id)

    # --- Configuration Methods
    def load_client_from_id(self, config_id):
        """Load remote server from configuration id."""
        options = self.load_conf(config_id)
        self.load_client(config_id, options)

    def load_client(self, config_id: str, options: SSHClientOptions):
        """Load remote server."""
        client = SpyderRemoteClient(config_id, options, _plugin=self)
        self._remote_clients[config_id] = client

    def load_conf(self, config_id):
        """Load remote server configuration."""
        options = self.get_conf(self.CONF_SECTION_SERVERS, {}).get(
            config_id, {}
        )

        # We couldn't find saved options for config_id
        if not options:
            return {}

        if options["client_keys"]:
            passpharse = self.get_conf(f"{config_id}/passpharse", secure=True)
            options["client_keys"] = [options["client_keys"]]

            # Passphrase is optional
            if passpharse:
                options["passpharse"] = passpharse
        elif options["config"]:
            # TODO: Check how this needs to be handled
            pass
        else:
            # Password is mandatory in this case
            password = self.get_conf(f"{config_id}/password", secure=True)
            options["password"] = password

        # Default for now
        options["platform"] = "linux"

        return SSHClientOptions(**options)

    def get_loaded_servers(self):
        """Get configured remote servers."""
        return self._remote_clients.keys()

    def get_config_ids(self):
        """Get configured remote servers ids."""
        return self.get_conf(self.CONF_SECTION_SERVERS, {}).keys()

    @Slot(str)
    def create_ipyclient_for_server(self, config_id):
        """Create a new IPython console client for a server."""
        auth_method = self.get_conf(f"{config_id}/auth_method")
        hostname = self.get_conf(f"{config_id}/{auth_method}/name")

        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        ipyclient = ipyconsole.create_client_for_kernel(
            # The connection file will be supplied when connecting a remote
            # kernel to this client
            connection_file="",
            # We use the server name as hostname because for clients it's the
            # attribute used by the IPython console to set their tab name.
            hostname=hostname,
            # These values are not necessary at this point.
            sshkey=None,
            password=None,
            # We save the server id in the client to perform on it operations
            # related to this plugin.
            server_id=config_id,
            # This is necessary because it takes a while before getting a
            # response from the server with the kernel id that will be
            # associated to this client. So, if users could close it before
            # that then it'll not be possible to shutdown that kernel unless
            # the server is stopped as well.
            can_close=False,
        )

        # IMPORTANT NOTE: We use a signal here instead of calling directly
        # container.on_kernel_started because doing that generates segfaults
        # and odd issues (e.g. the Variable Explorer not working).
        future = self._start_new_kernel(config_id)
        future.add_done_callback(
            lambda future: self._sig_kernel_started.emit(
                ipyclient, future.result()
            )
        )

    # ---- Private API
    # -------------------------------------------------------------------------
    # --- Remote Server Kernel Methods
    @Slot(str)
    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def get_kernels(self, config_id) -> list:
        """Get opened kernels."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            return await client.list_kernels()
        return []

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _get_kernel_info(self, config_id, kernel_id) -> dict:
        """Get kernel info."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            return await client.get_kernel_info_ensure_server(kernel_id)
        return {}

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _shutdown_kernel(self, config_id, kernel_id):
        """Shutdown a running kernel."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            with contextlib.suppress(Exception):
                await client.terminate_kernel(kernel_id)

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _start_new_kernel(self, config_id):
        """Start new kernel."""
        if config_id not in self._remote_clients:
            self.load_client_from_id(config_id)

        client = self._remote_clients[config_id]
        return await client.start_new_kernel_ensure_server()

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _restart_kernel(self, config_id, kernel_id) -> bool:
        """Restart kernel."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            with contextlib.suppress(Exception):
                return await client.restart_kernel(kernel_id)

        return False

    @AsyncDispatcher.dispatch(loop="asyncssh")
    async def _interrupt_kernel(self, config_id, kernel_id):
        """Interrupt kernel."""
        if config_id in self._remote_clients:
            client = self._remote_clients[config_id]
            await client.interrupt_kernel(kernel_id)

    def _reset_status(self):
        """Reset status of servers."""
        for config_id in self.get_config_ids():
            self.set_conf(f"{config_id}/status", ConnectionStatus.Inactive)
            self.set_conf(f"{config_id}/status_message", "")

    def _add_remote_consoles_menu(self):
        """Add remote consoles submenu to the Consoles menu."""
        container = self.get_container()
        container.setup_remote_consoles_submenu(render=False)

        menu = container.get_menu(RemoteClientMenus.RemoteConsoles)
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.add_item_to_application_menu(
            menu,
            menu_id=ApplicationMenus.Consoles,
            section=ConsolesMenuSections.New,
            before=IPythonConsoleWidgetActions.ConnectToKernel,
        )

        self._is_consoles_menu_added = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


"""
Bootstrap Spyder.

Execute Spyder from source checkout.
"""

# Standard library imports
import argparse
import os
import shutil
import subprocess
import sys
import time
from logging import Formatter, StreamHandler, getLogger
from pathlib import Path

# Local imports
from install_dev_repos import DEVPATH, REPOS, install_repo


# =============================================================================
# ---- Setup logger
# =============================================================================
fmt = Formatter('%(asctime)s [%(levelname)s] [%(name)s] -> %(message)s')
h = StreamHandler()
h.setFormatter(fmt)
logger = getLogger('Bootstrap')
logger.addHandler(h)
logger.setLevel('INFO')

time_start = time.time()


# =============================================================================
# ---- Parse command line
# =============================================================================
parser = argparse.ArgumentParser(
    usage="python bootstrap.py [options] [-- spyder_options]",
    epilog="""\
Arguments for Spyder's main script are specified after the --
symbol (example: `python bootstrap.py -- --hide-console`).
Type `python bootstrap.py -- --help` to read about Spyder
options.""")
parser.add_argument('--gui', default=None,
                    help="GUI toolkit: pyqt5, pyside2, pyqt6 or pyside6")
parser.add_argument('--hide-console', action='store_true', default=False,
                    help="Hide parent console window (Windows only)")
parser.add_argument('--safe-mode', dest="safe_mode",
                    action='store_true', default=False,
                    help="Start Spyder with a clean configuration directory")
parser.add_argument('--debug', action='store_true',
                    default=False, help="Run Spyder in debug mode")
parser.add_argument('--filter-log', default='',
                    help="Comma-separated module name hierarchies whose log "
                         "messages should be shown. e.g., "
                         "spyder.plugins.completion,spyder.plugins.editor")
parser.add_argument('--no-install', action='store_true', default=False,
                    help="Do not install Spyder or its subrepos")
parser.add_argument('spyder_options', nargs='*')

args = parser.parse_args()

assert args.gui in (None, "pyqt5", "pyside2", "pyqt6", "pyside6"), (
    "Invalid GUI toolkit option '%s'" % args.gui
)


# =============================================================================
# ---- Install sub repos
# =============================================================================
installed_dev_repo = False
if not args.no_install:
    prev_branch = None
    boot_branch_file = DEVPATH / ".boot_branch.txt"
    if boot_branch_file.exists():
        prev_branch = boot_branch_file.read_text()

    result = subprocess.run(
        ["git", "-C", DEVPATH, "merge-base", "--fork-point", "master"],
        capture_output=True
    )
    branch = "master" if result.stdout else "not master"
    boot_branch_file.write_text(branch)

    logger.info("Previous root branch: %s; current root branch: %s",
                prev_branch, branch)

    if branch != prev_branch:
        logger.info("Detected root branch change to/from master. "
                    "Will reinstall Spyder in editable mode.")
        REPOS[DEVPATH.name]["editable"] = False

    for name in REPOS.keys():
        # Don't install the spyder-remote-services subrepo because it's not
        # necessary on the Spyder side.
        if name != "spyder-remote-services":
            if not REPOS[name]['editable']:
                install_repo(name)
                installed_dev_repo = True
            else:
                logger.info("%s installed in editable mode", name)

if installed_dev_repo:
    logger.info("Restarting bootstrap to pick up installed subrepos")
    if '--' in sys.argv:
        sys.argv.insert(sys.argv.index('--'), '--no-install')
    else:
        sys.argv.append('--no-install')
    result = subprocess.run([sys.executable, *sys.argv])
    sys.exit(result.returncode)

# Local imports
# Must follow install_repo in case Spyder was not originally installed.
from spyder import get_versions

logger.info("Executing Spyder from source checkout")

# Prepare arguments for Spyder's main script
original_sys_argv = sys.argv.copy()
sys.argv = [sys.argv[0]] + args.spyder_options


# =============================================================================
# ---- Update os.environ
# =============================================================================
# Store variable to be used in self.restart (restart Spyder instance)
os.environ['SPYDER_BOOTSTRAP_ARGS'] = str(original_sys_argv[1:])

# Start Spyder with a clean configuration directory for testing purposes
if args.safe_mode:
    os.environ['SPYDER_SAFE_MODE'] = 'True'

# To activate/deactivate certain things for development
os.environ['SPYDER_DEV'] = 'True'

if args.debug:
    # safety check - Spyder config should not be imported at this point
    if "spyder.config.base" in sys.modules:
        sys.exit("ERROR: Can't enable debug mode - Spyder is already imported")
    logger.info("Switching debug mode on")
    os.environ["SPYDER_DEBUG"] = "3"
    if len(args.filter_log) > 0:
        logger.info("Displaying log messages only from the "
                    "following modules: %s", args.filter_log)
    os.environ["SPYDER_FILTER_LOG"] = args.filter_log
    # this way of interaction suxx, because there is no feedback
    # if operation is successful

# Selecting the GUI toolkit: PyQt5 if installed
if args.gui is None:
    try:
        import PyQt5  # analysis:ignore
        logger.info("PyQt5 is detected, selecting")
        os.environ['QT_API'] = 'pyqt5'
    except ImportError:
        sys.exit("ERROR: No PyQt5 detected!")
else:
    logger.info("Skipping GUI toolkit detection")
    os.environ['QT_API'] = args.gui


# =============================================================================
# ---- Check versions
# =============================================================================
# Checking versions (among other things, this has the effect of setting the
# QT_API environment variable if this has not yet been done just above)
versions = get_versions(reporev=True)
logger.info("Imported Spyder %s - Revision %s, Branch: %s; "
            "[Python %s %dbits, Qt %s, %s %s on %s]",
            versions['spyder'], versions['revision'], versions['branch'],
            versions['python'], versions['bitness'], versions['qt'],
            versions['qt_api'], versions['qt_api_ver'], versions['system'])

# Check that we have the right qtpy version
from spyder.utils import programs
if not programs.is_module_installed('qtpy', '>=1.1.0'):
    sys.exit("ERROR: Your qtpy version is outdated. Please install qtpy "
             "1.1.0 or higher to be able to work with Spyder!")


# =============================================================================
# ---- Execute Spyder
# =============================================================================
if args.hide_console and os.name == 'nt':
    logger.info("Hiding parent console (Windows only)")
    sys.argv.append("--hide-console")  # Windows only: show parent console

# Reset temporary config directory if starting in --safe-mode
if args.safe_mode or os.environ.get('SPYDER_SAFE_MODE'):
    from spyder.config.base import get_conf_path  # analysis:ignore
    conf_dir = Path(get_conf_path())
    if conf_dir.is_dir():
        shutil.rmtree(conf_dir)

logger.info("Running Spyder")
from spyder.app import start  # analysis:ignore

time_lapse = time.time() - time_start
logger.info("Bootstrap completed in %s%s",
            time.strftime("%H:%M:%S.", time.gmtime(time_lapse)),
            # gmtime() converts float into tuple, but loses milliseconds
            ("%.4f" % time_lapse).split('.')[1])

start.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder combobox widgets.

Use these widgets for any combobox you want to add to Spyder.
"""

# Standard library imports
import sys

# Third-party imports
import qstylizer.style
from qtpy import PYQT5, PYQT6
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QColor, QPainter, QFontMetrics
from qtpy.QtWidgets import (
    QComboBox,
    QFontComboBox,
    QFrame,
    QLineEdit,
    QListView,
    QProxyStyle,
    QStyle,
    QStyledItemDelegate,
    QStyleOptionFrame,
)
from superqt.utils import qdebounced

# Local imports
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, WIN


class _SpyderComboBoxProxyStyle(QProxyStyle):
    """Style proxy to adjust qdarkstyle issues."""

    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QStyle.SH_ComboBox_Popup:
            # Disable combobox popup top & bottom areas.
            # See spyder-ide/spyder#9682.
            # Taken from https://stackoverflow.com/a/21019371
            return 0

        return QProxyStyle.styleHint(self, hint, option, widget, returnData)


class _SpyderComboBoxDelegate(QStyledItemDelegate):
    """
    Delegate to make separators color follow our theme.

    Adapted from https://stackoverflow.com/a/33464045/438386
    """

    def __init__(self, parent, elide_mode=None):
        super().__init__(parent)
        self._elide_mode = elide_mode

    def paint(self, painter, option, index):
        data = index.data(Qt.AccessibleDescriptionRole)
        if data and data == "separator":
            painter.setPen(QColor(SpyderPalette.COLOR_BACKGROUND_6))
            painter.drawLine(
                option.rect.left() + AppStyle.MarginSize,
                option.rect.center().y(),
                option.rect.right() - AppStyle.MarginSize,
                option.rect.center().y()
            )
            return

        if self._elide_mode is not None:
            option.textElideMode = self._elide_mode

        super().paint(painter, option, index)

    def sizeHint(self, option, index):
        data = index.data(Qt.AccessibleDescriptionRole)
        if data and data == "separator":
            return QSize(0, 3 * AppStyle.MarginSize)

        return super().sizeHint(option, index)


class _SpyderComboBoxLineEdit(QLineEdit):
    """Lineedit used for comboboxes."""

    sig_mouse_clicked = Signal()

    def __init__(self, parent, editable, elide_mode=None):
        super().__init__(parent)
        self._editable = editable
        self._elide_mode = elide_mode
        self._focus_in = False

        # Fix style issues
        css = qstylizer.style.StyleSheet()
        css.QLineEdit.setValues(
            # These are necessary on Windows to prevent some ugly visual
            # glitches.
            backgroundColor="transparent",
            border="none",
            padding="0px",
            # Make text look centered for short comboboxes
            paddingRight=f"-{3 if WIN else 2}px"
        )

        self.setStyleSheet(css.toString())

    def mouseReleaseEvent(self, event):
        if not self._editable:
            # Emit a signal to display the popup afterwards
            self.sig_mouse_clicked.emit()
        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        if not self._editable:
            # Avoid selecting the lineedit text with double clicks
            pass
        else:
            super().mouseDoubleClickEvent(event)

    def focusInEvent(self, event):
        self._focus_in = True
        super().focusInEvent(event)

    def focusOutEvent(self, event):
        self._focus_in = False
        super().focusOutEvent(event)

    def paintEvent(self, event):
        if self._elide_mode is not None and not self._focus_in:
            # This code is taken for the most part from the
            # AmountEdit.paintEvent method, part of the Electrum project. See
            # the Electrum entry in our NOTICE.txt file for the details.
            # Licensed under the MIT license.
            painter = QPainter(self)
            option = QStyleOptionFrame()
            self.initStyleOption(option)

            text_rect = self.style().subElementRect(
                QStyle.SE_LineEditContents, option, self
            )

            # Neded so the text is placed correctly according to our style
            text_rect.adjust(2, 0, 0, 0)

            fm = QFontMetrics(self.font())
            text = fm.elidedText(
                self.text(), self._elide_mode, text_rect.width()
            )

            color = (
                SpyderPalette.COLOR_TEXT_1
                if self.isEnabled()
                else SpyderPalette.COLOR_DISABLED
            )
            painter.setPen(QColor(color))
            painter.drawText(
                text_rect, int(Qt.AlignLeft | Qt.AlignVCenter), text
            )

            return

        super().paintEvent(event)


class _SpyderComboBoxView(QListView):
    """Listview used for comboboxes"""

    sig_current_item_changed = Signal(object)

    def currentChanged(self, current, previous):
        # This covers selecting a different item with the keyboard or when
        # hovering with the mouse over the list.
        self.sig_current_item_changed.emit(current)
        super().currentChanged(current, previous)


class _SpyderComboBoxMixin:
    """Mixin with the basic style and functionality for our comboboxes."""

    sig_item_in_popup_changed = Signal(str)
    """
    This signal is emitted when an item in the combobox popup (i.e. dropdown)
    has changed.

    Parameters
    ----------
    item: str
        Item text
    """

    sig_popup_is_hidden = Signal()
    """
    This signal is emitted when the combobox popup (i.e. dropdown) is hidden.
    """

    def __init__(self):

        # Style
        self._css = self._generate_stylesheet()
        self.setStyleSheet(self._css.toString())

        style = _SpyderComboBoxProxyStyle(None)
        style.setParent(self)
        self.setStyle(style)

        # Report when the current item in the dropdown has changed
        view = _SpyderComboBoxView(self)
        view.sig_current_item_changed.connect(self._on_item_changed)
        self.setView(view)

    def contextMenuEvent(self, event):
        # Prevent showing context menu for editable comboboxes because it's
        # added automatically by Qt. That means the menu is not built using our
        # API and it's not localized.
        pass

    @qdebounced(timeout=100)
    def _on_item_changed(self, index):
        if index.isValid():
            self.sig_item_in_popup_changed.emit(index.data())

    def _generate_stylesheet(self):
        """Base stylesheet for Spyder comboboxes."""
        css = qstylizer.style.StyleSheet()

        # Make our comboboxes have a uniform height
        css.QComboBox.setValues(
            minHeight=f'{AppStyle.ComboBoxMinHeight}em'
        )

        # Add top and bottom padding to the inner contents of comboboxes
        css["QComboBox QAbstractItemView"].setValues(
            paddingTop=f"{AppStyle.MarginSize + 1}px",
            paddingBottom=f"{AppStyle.MarginSize + 1}px"
        )

        # Add margin and padding to combobox items
        css["QComboBox QAbstractItemView::item"].setValues(
            marginLeft=f"{AppStyle.MarginSize}px",
            marginRight=f"{AppStyle.MarginSize}px",
            padding=f"{AppStyle.MarginSize}px"
        )

        # Make color of hovered combobox items match the one used in other
        # Spyder widgets
        css["QComboBox QAbstractItemView::item:selected:active"].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_3,
        )

        return css


class SpyderComboBox(QComboBox, _SpyderComboBoxMixin):
    """Default combobox widget for Spyder."""

    def __init__(self, parent=None, items_elide_mode=None):
        """
        Default combobox widget for Spyder.

        Parameters
        ----------
        parent: QWidget, optional
            The combobox parent.
        items_elide_mode: Qt.TextElideMode, optional
            Elide mode for the combobox items.
        """
        if PYQT5 or PYQT6:
            super().__init__(parent)
        else:
            QComboBox.__init__(self, parent)
            _SpyderComboBoxMixin.__init__(self)

        self.is_editable = None
        self._is_shown = False
        self._is_popup_shown = False

        # This is necessary to have more fine-grained control over the style of
        # our comboboxes with css, e.g. to add more padding between its items.
        # See https://stackoverflow.com/a/33464045/438386 for the details.
        self.setItemDelegate(
            _SpyderComboBoxDelegate(self, elide_mode=items_elide_mode)
        )

    def showEvent(self, event):
        """Adjustments when the widget is shown."""

        if not self._is_shown:
            if not self.isEditable():
                self.is_editable = False
                self.setLineEdit(_SpyderComboBoxLineEdit(self, editable=False))

                # This is necessary to make Qt position the popup widget below
                # the combobox for non-editable ones.
                # Solution from https://stackoverflow.com/a/45191141/438386
                self.setEditable(True)
                self.lineEdit().setReadOnly(True)

                # Show popup when the lineEdit is clicked, which is the default
                # behavior for non-editable comboboxes in Qt.
                self.lineEdit().sig_mouse_clicked.connect(self.showPopup)
            else:
                self.is_editable = True

            self._is_shown = True

        super().showEvent(event)

    def showPopup(self):
        """Adjustments when the popup is shown."""
        super().showPopup()

        if sys.platform == "darwin":
            # Reposition popup to display it in the right place.
            # Solution from https://forum.qt.io/post/349517
            popup = self.findChild(QFrame)
            popup.move(popup.x() - 3, popup.y() + 4)

            # Adjust width to match the lineEdit one.
            if not self._is_popup_shown:
                popup.setFixedWidth(popup.width() + 2)
                self._is_popup_shown = True
        else:
            # Make borders straight to make popup feel as part of the combobox.
            # This doesn't work reliably on Mac.
            self._css.QComboBox.setValues(
                borderBottomLeftRadius="0px",
                borderBottomRightRadius="0px",
            )

            self.setStyleSheet(self._css.toString())

    def hidePopup(self):
        """Adjustments when the popup is hidden."""
        super().hidePopup()
        self.sig_popup_is_hidden.emit()

        if not sys.platform == "darwin":
            # Make borders rounded when popup is not visible. This doesn't work
            # reliably on Mac.
            self._css.QComboBox.setValues(
                borderBottomLeftRadius=SpyderPalette.SIZE_BORDER_RADIUS,
                borderBottomRightRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            )

            self.setStyleSheet(self._css.toString())


class SpyderComboBoxWithIcons(SpyderComboBox):
    """"Combobox widget for Spyder when its items have icons."""

    def __init__(self, parent=None, items_elide_mode=None):
        """"
        Combobox widget for Spyder when its items have icons.

        Parameters
        ----------
        parent: QWidget, optional
            The combobox parent.
        items_elide_mode: Qt.TextElideMode, optional
            Elide mode for the combobox items.
        """
        super().__init__(parent, items_elide_mode)

        # Padding is not necessary because icons give items enough of it.
        self._css["QComboBox QAbstractItemView::item"].setValues(
            padding="0px"
        )

        self.setStyleSheet(self._css.toString())


class SpyderFontComboBox(QFontComboBox, _SpyderComboBoxMixin):

    def __init__(self, parent=None):
        if PYQT5 or PYQT6:
            super().__init__(parent)
        else:
            QFontComboBox.__init__(self, parent)
            _SpyderComboBoxMixin.__init__(self)

        # Avoid font name eliding because it confuses users.
        # Fixes spyder-ide/spyder#22683
        self.setItemDelegate(
            _SpyderComboBoxDelegate(self, elide_mode=Qt.ElideNone)
        )

        # Elide selected font name in case it's too long
        self.setLineEdit(
            _SpyderComboBoxLineEdit(
                self, editable=True, elide_mode=Qt.ElideMiddle
            )
        )

        # Adjust popup width to contents.
        self.setSizeAdjustPolicy(
            QComboBox.AdjustToMinimumContentsLengthWithIcon
        )

    def showPopup(self):
        """Adjustments when the popup is shown."""
        super().showPopup()

        if sys.platform == "darwin":
            popup = self.findChild(QFrame)
            popup.move(popup.x() - 3, popup.y() + 4)

    def hidePopup(self):
        super().hidePopup()
        self.sig_popup_is_hidden.emit()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Fixture to create files folders.
"""

# Standard imports
import os
import os.path as osp
import sys

# Third party imports
import pytest

# Local imports
from spyder.py3compat import to_text_string
from spyder.plugins.explorer.widgets.fileassociations import (
    FileAssociationsWidget)


@pytest.fixture(params=[
        ['script.py', 'dir1/dir2/dir3/dir4'],
        ['script.py', 'script1.py', 'testdir/dir1/script2.py'],
        ['subdir/innerdir/dir3/text.txt', 'dir1/dir2/dir3',
         'dir1/dir2/dir3/file.txt', 'dir1/dir2/dir3/dir4/python.py'],
    ]
)
def create_folders_files(tmpdir, request):
    """A project directory with dirs and files for testing."""
    project_dir = to_text_string(tmpdir.mkdir('project'))
    destination_dir = to_text_string(tmpdir.mkdir('destination'))
    top_folder = osp.join(project_dir, 'top_folder_in_proj')
    if not osp.exists(top_folder):
        os.mkdir(top_folder)
    list_paths = []
    for item in request.param:
        if osp.splitext(item)[1]:
            if osp.split(item)[0]:
                dirs, fname = osp.split(item)
                dirpath = osp.join(top_folder, dirs)
                if not osp.exists(dirpath):
                    os.makedirs(dirpath)
                    item_path = osp.join(dirpath, fname)
            else:
                item_path = osp.join(top_folder, item)
        else:
            dirpath = osp.join(top_folder, item)
            if not osp.exists(dirpath):
                os.makedirs(dirpath)
                item_path = dirpath
        if not osp.isdir(item_path):
            with open(item_path, 'w') as fh:
                fh.write("File Path:\n" + str(item_path).replace(os.sep, '/'))
        list_paths.append(item_path)
    return list_paths, project_dir, destination_dir, top_folder


@pytest.fixture
def file_assoc_widget(qtbot, tmp_path):
    widget = FileAssociationsWidget()
    qtbot.addWidget(widget)
    if os.name == 'nt':
        ext = '.exe'
        path_obj = tmp_path / ('app 2' + ext)
        path_obj.write_bytes(b'Binary file contents')
        fpath = str(path_obj)
    elif sys.platform == 'darwin':
        ext = '.app'
        path_obj = tmp_path / ('app 2' + ext)
        path_obj.mkdir()
        fpath = str(path_obj)
    else:
        ext = '.desktop'
        path_obj = tmp_path / ('app 2' + ext)
        path_obj.write_text(u'Text file contents')
        fpath = str(path_obj)

    data = {
        '*.csv':
            [
                ('App name 1', '/path/to/app 1' + ext),
                ('App name 2', fpath),
            ],
        '*.txt':
            [
                ('App name 2', fpath),
                ('App name 3', '/path/to/app 3' + ext),
            ],
    }
    widget.load_values(data)
    widget.show()
    widget.test_data = data
    return qtbot, widget



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder run container."""

# Standard library imports
import functools
import os.path as osp
from typing import Callable, List, Dict, Tuple, Set, Optional
from uuid import uuid4
from weakref import WeakSet, WeakValueDictionary

# Third-party imports
from qtpy.QtGui import QIcon
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QAction

# Local imports
from spyder.api.plugins import Plugins
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.api.translations import _
from spyder.plugins.run.api import (
    RunActions, StoredRunExecutorParameters, RunContext, RunExecutor,
    RunResultFormat, RunConfigurationProvider, RunResultViewer, OutputFormat,
    SupportedExecutionRunConfiguration, RunConfigurationMetadata,
    StoredRunConfigurationExecutor, ExtendedRunExecutionParameters,
    RunExecutionParameters, WorkingDirOpts, WorkingDirSource,
    SupportedExtensionContexts)
from spyder.plugins.run.models import (
    RunExecutorParameters, RunExecutorListModel, RunConfigurationListModel)
from spyder.plugins.run.widgets import RunDialog, RunDialogStatus
from spyder.utils.sourcecode import camel_case_to_snake_case


class RunContainer(PluginMainContainer):
    """Non-graphical container used to spawn dialogs and creating actions."""

    sig_run_action_created = Signal(str, bool, str)
    sig_open_preferences_requested = Signal()

    # ---- PluginMainContainer API
    # -------------------------------------------------------------------------
    def setup(self):
        self.current_working_dir: Optional[str] = None

        self.parameter_model = RunExecutorParameters(self)
        self.executor_model = RunExecutorListModel(self)
        self.metadata_model = RunConfigurationListModel(
            self, self.executor_model)

        self.run_metadata_provider: WeakValueDictionary[
            str, RunConfigurationProvider] = WeakValueDictionary()
        self.run_executors: WeakValueDictionary[
            str, RunExecutor] = WeakValueDictionary()
        self.viewers: WeakValueDictionary[
            str, RunResultViewer] = WeakValueDictionary()

        self.executor_use_count: Dict[str, int] = {}
        self.viewers_per_output: Dict[str, Set[str]] = {}

        self.run_action = self.create_action(
            RunActions.Run,
            _('&Run'),
            self.create_icon('run'),
            tip=_("Run file"),
            triggered=functools.partial(
                self.run_file, selected_uuid=None,
                selected_executor=None
            ),
            register_shortcut=True,
            shortcut_context='_',
            context=Qt.ApplicationShortcut
        )

        self.configure_action = self.create_action(
            RunActions.Configure,
            _('&Configuration per file'),
            self.create_icon('run_settings'),
            tip=_('Run settings'),
            triggered=functools.partial(
                self.edit_run_configurations,
                display_dialog=True,
                disable_run_btn=True,
                selected_executor=None
            ),
            register_shortcut=True,
            shortcut_context='_',
            context=Qt.ApplicationShortcut
        )

        self.create_action(
            RunActions.GlobalConfigurations,
            _("&Global configurations"),
            triggered=self.sig_open_preferences_requested
        )

        self.re_run_action = self.create_action(
            RunActions.ReRun,
            _('Re-run &last file'),
            self.create_icon('run_again'),
            tip=_('Run again last file'),
            triggered=self.re_run_file,
            register_shortcut=True,
            shortcut_context='_',
            context=Qt.ApplicationShortcut
        )

        self.re_run_action.setEnabled(False)

        self.current_input_provider: Optional[str] = None
        self.current_input_extension: Optional[str] = None

        self.context_actions: Dict[
            Tuple[str, str, str], Tuple[QAction, Callable]] = {}
        self.re_run_actions: Dict[
            Tuple[str, str, str], Tuple[QAction, Callable]] = {}
        self.run_executor_actions: Dict[
            Tuple[str, str], Tuple[QAction, Callable]] = {}

        # 'File' is the only super context we support for now. It needs to be
        # set here so that run actions for files in different plugins (e.g.
        # 'Debug file') are declared correctly at startup.
        # Fixes spyder-ide/spyder#23694
        self.super_contexts: Set[str] = {RunContext.File}
        self.supported_extension_contexts: Dict[str, Set[Tuple[str, str]]] = {}

        self.last_executed_file: Optional[str] = None
        self.last_executed_per_context: Set[Tuple[str, str]] = set()
        self._last_executed_configuration: Optional[str] = None
        self._save_last_configured_executor: bool = True

    def update_actions(self):
        pass

    # ---- Public API
    # -------------------------------------------------------------------------
    def gen_anonymous_execution_run(
        self,
        context: str,
        extra_action_name: Optional[str] = None,
        context_modificator: Optional[str] = None,
        re_run: bool = False,
        last_executor_name: Optional[str] = None
    ) -> Callable:

        def anonymous_execution_run():
            if self.currently_selected_configuration is None:
                return

            # We save the last executed configuration in case we need it to
            # re-run the last action with the same parameters.
            # This is necessary for spyder-ide/spyder#23076
            if not re_run:
                uuid = self.currently_selected_configuration
                self._last_executed_configuration = (
                    self.currently_selected_configuration
                )
            else:
                uuid = self._last_executed_configuration

                # If nothing has been executed so far, we can't re-run it.
                if uuid is None:
                    return

            # Get run configuration
            input_provider = self.run_metadata_provider[uuid]
            if context in self.super_contexts:
                run_conf = input_provider.get_run_configuration(uuid)
            else:
                run_conf = input_provider.get_run_configuration_per_context(
                    context,
                    extra_action_name,
                    context_modificator,
                    re_run=re_run,
                )

            if run_conf is None:
                return

            # Get last executor
            super_metadata = self.metadata_model[uuid]
            extension = super_metadata['input_extension']

            last_executor = last_executor_name or self._get_default_executor(
                extension
            )
            if last_executor is None:
                last_executor = self.get_last_used_executor_parameters(uuid)
                last_executor = last_executor['executor']

            run_comb = (extension, context)
            if (
                last_executor is None
                or not self.executor_model.executor_supports_configuration(
                    last_executor, run_comb
                )
            ):
                last_executor = self.executor_model.get_default_executor(
                    run_comb)

            # Get execution params
            if context in self.super_contexts:
                # This applies to super contexts because we save run conf
                # params only for them (we handle just one super context for
                # now: "file").

                # Get last used params
                all_params = (
                    self.metadata_model.get_run_configuration_parameters(
                        uuid, last_executor
                    )["params"]
                )

                last_params_uuid = self.get_last_used_execution_params(
                    uuid, last_executor
                )

                if last_params_uuid in all_params:
                    # Use the last params set by the user in RunDialog.
                    # Fixes spyder-ide/spyder#22496
                    exec_params = all_params[last_params_uuid]["params"]
                else:
                    # If no custom params have been set, use the default ones
                    for stored_params in all_params.values():
                        if stored_params.get("default", False):
                            exec_params = stored_params["params"]
            else:
                # This is necessary for subcontexts (e.g. cell and selection).
                # Although they can have conf widgets (which is also tested),
                # that functionality is not exposed in the UI at the moment.
                path = super_metadata['path']
                dirname = osp.dirname(path)

                executor_metadata = self.executor_model[
                    ((extension, context), last_executor)
                ]
                ConfWidget = executor_metadata['configuration_widget']

                conf = {}
                if ConfWidget is not None:
                    conf = ConfWidget.get_default_configuration()

                working_dir = WorkingDirOpts(
                    source=WorkingDirSource.ConfigurationDirectory,
                    path=dirname
                )

                exec_params = RunExecutionParameters(
                    working_dir=working_dir, executor_params=conf
                )

            # Perform execution
            ext_exec_params = ExtendedRunExecutionParameters(
                uuid=None, name=None, params=exec_params)
            executor = self.run_executors[last_executor]
            executor.exec_run_configuration(run_conf, ext_exec_params)

            # Save metadata and set state of re-run actions
            self.last_executed_per_context |= {(uuid, context)}

            if (
                (context, extra_action_name, context_modificator)
                in self.re_run_actions
            ):
                act, _ = self.re_run_actions[
                    (context, extra_action_name, context_modificator)]
                act.setEnabled(True)

        return anonymous_execution_run

    def run_file(self, selected_uuid=None, selected_executor=None):
        if not isinstance(selected_uuid, bool) and selected_uuid is not None:
            self.switch_focused_run_configuration(selected_uuid)

        self.edit_run_configurations(
            display_dialog=False,
            selected_executor=selected_executor
        )

    def edit_run_configurations(
        self,
        display_dialog=True,
        disable_run_btn=False,
        selected_executor=None
    ):
        self.dialog = RunDialog(
            self,
            self.metadata_model,
            self.executor_model,
            self.parameter_model,
            disable_run_btn=disable_run_btn
        )

        self.dialog.setup()
        self.dialog.finished.connect(self.process_run_dialog_result)
        self.dialog.sig_delete_config_requested.connect(
            self.delete_executor_configuration_parameters
        )

        # If the file is going to be executed (not configured) and the executor
        # is not provided, check if it has a default one.
        if not display_dialog and selected_executor is None:
            metadata = self.metadata_model.get_selected_metadata()
            if metadata is not None:
                extension = metadata["input_extension"]
                selected_executor = self._get_default_executor(extension)

        if selected_executor is not None:
            self.dialog.select_executor(selected_executor)

        if display_dialog:
            self.dialog.open()
        else:
            self.dialog.run_btn_clicked()

    def process_run_dialog_result(self, result):
        status = self.dialog.status

        if status == RunDialogStatus.Close:
            return

        uuid, executor_name, ext_params = self.dialog.get_configuration()

        if (status & RunDialogStatus.Save) == RunDialogStatus.Save:
            exec_uuid = ext_params['uuid']

            # Default parameters should already be saved in our config system.
            # So, there is no need to save them again here.
            if exec_uuid is not None and not ext_params["default"]:

                info = self.metadata_model.get_metadata_context_extension(uuid)
                context, ext =  info
                context_name = context['name']
                context_id = getattr(RunContext, context_name)
                all_exec_params = self.get_executor_configuration_parameters(
                    executor_name,
                    ext,
                    context_id
                )
                exec_params = all_exec_params['params']
                exec_params[exec_uuid] = ext_params

                self.set_executor_configuration_parameters(
                    executor_name,
                    ext,
                    context_id,
                    all_exec_params,
                )

            last_used_conf = StoredRunConfigurationExecutor(
                executor=executor_name, selected=ext_params['uuid']
            )

            if self._save_last_configured_executor:
                self._set_last_configured_executor(uuid, executor_name)

            # Reset this attribute to the default for next time
            self._save_last_configured_executor = True

            self.set_last_used_execution_params(uuid, last_used_conf)

        if (status & RunDialogStatus.Run) == RunDialogStatus.Run:
            provider = self.run_metadata_provider[uuid]
            executor = self.run_executors[executor_name]
            run_conf = provider.get_run_configuration(uuid)
            working_dir_opts = ext_params['params']['working_dir']
            working_dir_source = working_dir_opts['source']
            if working_dir_source == WorkingDirSource.ConfigurationDirectory:
                fname = run_conf['metadata']['path']
                dirname = osp.dirname(fname)
            elif working_dir_source == WorkingDirSource.CurrentDirectory:
                dirname = self.current_working_dir
            else:
                dirname = working_dir_opts['path']

            working_dir_opts['path'] = dirname

            self.last_executed_file = uuid
            self.re_run_action.setEnabled(True)

            executor.exec_run_configuration(run_conf, ext_params)

    def re_run_file(self):
        self.run_file(self.last_executed_file, selected_executor=None)
    
    @property
    def currently_selected_configuration(self):
        return self.metadata_model.get_current_run_configuration()

    def switch_focused_run_configuration(self, uuid: Optional[str]):
        uuid = uuid or None

        # We need the first check to correctly update the run and context
        # actions when the config is None.
        # Fixes spyder-ide/spyder#22607
        if uuid is not None and uuid == self.currently_selected_configuration:
            return

        self.metadata_model.set_current_run_configuration(uuid)

        if uuid is not None:
            self.run_action.setEnabled(True)
            self.configure_action.setEnabled(True)

            metadata = self.metadata_model[uuid]
            self.current_input_provider = metadata['source']
            self.current_input_extension = metadata['input_extension']

            input_provider = self.run_metadata_provider[uuid]
            input_provider.focus_run_configuration(uuid)
            self.set_actions_status()

            return

        self.run_action.setEnabled(False)
        self.configure_action.setEnabled(False)

        for context, act, mod in self.context_actions:
            action, __ = self.context_actions[(context, act, mod)]
            action.setEnabled(False)

        for context, act, mod in self.re_run_actions:
            action, __ = self.re_run_actions[(context, act, mod)]
            action.setEnabled(False)

        for context_name, executor_name in self.run_executor_actions:
            action, __ = self.run_executor_actions[
                (context_name, executor_name)]
            action.setEnabled(False)

    def set_actions_status(self):
        if self.current_input_provider is None:
            return

        if self.currently_selected_configuration is None:
            return

        if (
            self.current_input_provider
            not in self.supported_extension_contexts
        ):
            return

        input_provider_ext_ctxs = self.supported_extension_contexts[
            self.current_input_provider]

        for context, act, mod in self.context_actions:
            key = (self.current_input_extension, context)
            status = key in self.executor_model
            status = status and key in input_provider_ext_ctxs
            action, __ = self.context_actions[(context, act, mod)]
            action.setEnabled(status)

        for context, act, mod in self.re_run_actions:
            key = (self.current_input_extension, context)
            status = key in self.executor_model
            status = status and key in input_provider_ext_ctxs

            last_run_exists = (
                (self.currently_selected_configuration,
                 context) in self.last_executed_per_context)

            action, __ = self.re_run_actions[(context, act, mod)]
            action.setEnabled(status and last_run_exists)

        for context_name, executor_name in self.run_executor_actions:
            status = self.executor_model.executor_supports_configuration(
                executor_name,
                (self.current_input_extension, context_name))
            key = (self.current_input_extension, context_name)
            status = status and key in input_provider_ext_ctxs

            action, __ = self.run_executor_actions[
                (context_name, executor_name)]
            action.setEnabled(status)

    def set_current_working_dir(self, path: str):
        self.current_working_dir = path

    def create_run_button(
        self,
        context_name: str,
        text: str,
        icon: Optional[QIcon] = None,
        tip: Optional[str] = None,
        shortcut_context: Optional[str] = None,
        register_shortcut: bool = False,
        extra_action_name: Optional[str] = None,
        context_modificator: Optional[str] = None,
        re_run: bool = False
    ) -> QAction:
        """
        Create a run or a "run and do something" button for a specific run
        context.

        Parameters
        ----------
        context_name: str
            The identifier of the run context.
        text: str
           Localized text for the action
        icon: Optional[QIcon]
            Icon for the action when used in menu or toolbar.
        tip: Optional[str]
            Tooltip to define for action on menu or toolbar.
        shortcut_context: Optional[str]
            Set the `str` context of the shortcut.
        register_shortcut: bool
            If True, main window will expose the shortcut in Preferences.
            The default value is `False`.
        extra_action_name: Optional[str]
            The name of the action to execute on the run input provider
            after requesting the run input.
        context_modificator: Optional[str]
            The name of the modification to apply to the action, e.g. run
            selection <up to line>.
        re_run: bool
            If True, then the button will act as a re-run button instead of
            a run one.

        Returns
        -------
        action: SpyderAction
            The corresponding action that was created.

        Notes
        -----
        1. The context passed as a parameter must be a subordinate of the
        context of the current focused run configuration that was
        registered via `register_run_configuration_metadata`. For instance,
        Cell can be used if and only if the file was registered.

        2. The button will be registered as `run <context>` or
        `run <context> <context_modificator> and <extra_action_name>` 
        on the action registry.

        3. The created button will operate over the last focused run input
        provider.

        4. If the requested button already exists, this method will not do
        anything, which implies that the first registered shortcut will be the
        one to be used. For the built-in run contexts (file, cell and
        selection), the editor will register their corresponding icons and
        shortcuts.
        """
        dict_actions = self.re_run_actions if re_run else self.context_actions

        if (
            (context_name, extra_action_name, context_modificator)
            in dict_actions
        ):
            action, __ = self.context_actions[
                (context_name, extra_action_name, context_modificator)]
            return action

        prefix = 're-' if re_run else ''
        action_name = f'{prefix}run {context_name}'
        if context_modificator is not None:
            action_name = f'{action_name} {context_modificator}'
        if extra_action_name is not None:
            action_name = f'{action_name} and {extra_action_name}'

        func = self.gen_anonymous_execution_run(
            context_name, extra_action_name, context_modificator,
            re_run=re_run, last_executor_name=None)

        action = self.create_action(
            action_name,
            text,
            icon,
            tip=tip,
            triggered=func,
            register_shortcut=register_shortcut,
            shortcut_context=shortcut_context,
            context=Qt.WidgetShortcut,
        )

        if re_run:
            self.re_run_actions[
                (context_name, extra_action_name, context_modificator)] = (
                action, func)
            action.setEnabled(False)
        else:
            self.context_actions[
                (context_name, extra_action_name, context_modificator)] = (
                action, func)

        self.sig_run_action_created.emit(action_name, register_shortcut,
                                         shortcut_context)
        return action

    def create_run_in_executor_button(
        self,
        context_name: str,
        executor_name: str,
        text: str,
        icon: Optional[QIcon] = None,
        tip: Optional[str] = None,
        shortcut_context: Optional[str] = None,
        register_shortcut: bool = False,
        shortcut_widget_context: int = Qt.WidgetShortcut
    ) -> QAction:
        """
        Create a "run <context> in <provider>" button for a given run context
        and executor.

        Parameters
        ----------
        context_name: str
            The identifier of the run context.
        executor_name: str
            The identifier of the run executor.
        text: str
           Localized text for the action
        icon: Optional[QIcon]
            Icon for the action when used in a menu or toolbar.
        tip: Optional[str]
            Tooltip to define for action on menu or toolbar.
        shortcut_context: Optional[str]
            Set the `str` context of the shortcut.
        register_shortcut: bool
            If True, main window will expose the shortcut in Preferences.
            The default value is `False`.

        Returns
        -------
        action: SpyderAction
            The corresponding action that was created.

        Notes
        -----
        1. The context passed as a parameter must be a subordinate of the
        context of the current focused run configuration that was
        registered via `register_run_configuration_metadata`. For instance,
        Cell can be used if and only if the file was registered.

        2. The button will be registered as `run <context> in <provider>` on
        the action registry.

        3. The created button will operate over the last focused run input
        provider.

        4. If the requested button already exists, this method will not do
        anything, which implies that the first registered shortcut will be the
        one to be used.
        """
        if (context_name, executor_name) in self.run_executor_actions:
            action, __ = self.run_executor_actions[
                (context_name, executor_name)]
            return action

        action_name = f'run {context_name} in {executor_name}'

        func = lambda: None
        if context_name in self.super_contexts:
            func = functools.partial(self.run_file,
                                     selected_executor=executor_name)
        else:
            func = self.gen_anonymous_execution_run(
                context_name, re_run=False,
                last_executor_name=executor_name)

        action = self.create_action(
            action_name,
            text,
            icon,
            tip=tip,
            triggered=func,
            register_shortcut=register_shortcut,
            shortcut_context=shortcut_context,
            context=shortcut_widget_context
        )

        self.run_executor_actions[(context_name, executor_name)] = (
            action, func)

        self.sig_run_action_created.emit(action_name, register_shortcut,
                                         shortcut_context)
        return action

    def register_run_configuration_provider(
        self,
        provider_name: str,
        supported_extensions_contexts: List[SupportedExtensionContexts]
    ):
        """
        Register the supported extensions and contexts that a
        `RunConfigurationProvider` supports.

        Parameters
        ----------
        provider_name: str
            The identifier of the :class:`RunConfigurationProvider` instance
            that is registering the set of supported contexts per extension.
        supported_extensions_contexts: List[SupportedExtensionContexts]
            A list containing the supported contexts per file extension
            supported.
        """
        provider_extensions_contexts = self.supported_extension_contexts.get(
            provider_name, set({}))

        for supported_extension_contexts in supported_extensions_contexts:
            for ext_context in supported_extension_contexts['contexts']:
                context = ext_context['context']
                is_super = ext_context['is_super']
                context_name = context['name']
                context_identifier = context.get('identifier', None)
                if context_identifier is None:
                    context_identifier = camel_case_to_snake_case(context_name)
                    context['identifier'] = context_identifier
                setattr(RunContext, context_name, context_identifier)
                ext_list = supported_extension_contexts['input_extension']
                if not isinstance(ext_list, list):
                    ext_list = [ext_list]
                for ext in ext_list:
                    provider_extensions_contexts |= {(ext, context_identifier)}
                if is_super:
                    self.super_contexts |= {context_identifier}

        self.supported_extension_contexts[
            provider_name] = provider_extensions_contexts

        self.set_actions_status()

    def deregister_run_configuration_provider(
        self,
        provider_name: str,
        unsupported_extensions_contexts: List[SupportedExtensionContexts]
    ):
        """
        Deregister the extensions and contexts that a
        `RunConfigurationProvider` no longer supports.

        Parameters
        ----------
        provider_name: str
            The identifier of the :class:`RunConfigurationProvider` instance
            that is deregistering the set of formerly supported contexts
            per extension.
        unsupported_extensions_contexts: List[SupportedExtensionContexts]
            A list containing the formerly supported contexts per
            file extension.
        """
        provider_extensions_contexts = self.supported_extension_contexts.get(
            provider_name, set({}))

        for unsupported_extension_contexts in unsupported_extensions_contexts:
            ext = unsupported_extension_contexts['input_extension']
            for ext_context in unsupported_extension_contexts['contexts']:
                context = ext_context['context']
                context_name = context['name']
                context_id = getattr(RunContext, context_name)
                provider_extensions_contexts -= {(ext, context_id)}

        if provider_extensions_contexts:
            self.supported_extension_contexts[
                provider_name] = provider_extensions_contexts
        else:
            self.supported_extension_contexts.pop(provider_name, set({}))

    def register_run_configuration_metadata(
        self,
        provider: RunConfigurationProvider,
        metadata: RunConfigurationMetadata
    ):
        """
        Register the metadata for a run configuration.

        Parameters
        ----------
        provider: RunConfigurationProvider
            A :class:`SpyderPluginV2` instance that implements the
            :class:`RunConfigurationProvider` interface and will register
            and own a run configuration.
        metadata: RunConfigurationMetadata
            The metadata for a run configuration that the provider is able to
            produce.
        """
        context = metadata['context']
        context_name = context['name']
        context_identifier = context.get('identifier', None)

        if context_identifier is None:
            context_identifier = camel_case_to_snake_case(context_name)
            context['identifier'] = context_identifier
        setattr(RunContext, context_name, context_identifier)

        run_id = metadata['uuid']
        self.run_metadata_provider[run_id] = provider
        self.metadata_model[run_id] = metadata

    def deregister_run_configuration_metadata(self, uuid: str):
        """
        Deregister a given run configuration by its unique identifier.

        Parameters
        ----------
        uuid: str
            Unique identifier for a run configuration metadata that will not
            longer exist. This id should have been registered using
            `register_run_configuration_metadata`.
        """
        self.metadata_model.pop(uuid)
        self.run_metadata_provider.pop(uuid)

        if uuid == self.last_executed_file:
            self.last_executed_file = None
            self.re_run_action.setEnabled(False)

        to_remove = set({})
        for (opened_uuid, context) in self.last_executed_per_context:
            if opened_uuid == uuid:
                to_remove |= {(opened_uuid, context)}

        self.last_executed_per_context -= to_remove

    def register_executor_configuration(
        self,
        executor: RunExecutor,
        configuration: List[SupportedExecutionRunConfiguration]
    ):
        """
        Register a :class:`RunExecutorProvider` instance to indicate its
        support for a given set of run configurations.

        Parameters
        ----------
        provider: RunExecutor
            A :class:`SpyderPluginV2` instance that implements the
            :class:`RunExecutor` interface and will register execution
            input type information.
        configuration: List[SupportedExecutionRunConfiguration]
            A list of input configurations that the provider is able to
            produce. Each configuration specifies the input extension
            identifier as well as the available execution context for that
            type.
        """
        executor_id = executor.NAME
        executor_name = executor.get_name()
        self.run_executors[executor_id] = executor
        executor_count = self.executor_use_count.get(executor_id, 0)

        for config in configuration:
            context = config['context']
            context_name = context['name']
            context_id = context.get('identifier', None)
            if context_id is None:
                context_id = camel_case_to_snake_case(context_name)
            setattr(RunContext, context_name, context_id)

            output_formats = []
            for out in config['output_formats']:
                output_name = out['name']
                output_id = out.get('identifier', None)
                if not output_id:
                    output_id = camel_case_to_snake_case(output_name)
                setattr(RunResultFormat, output_name, output_id)
                updated_out = {'name': output_name, 'identifier': output_id}
                output_formats.append(updated_out)

            config['output_formats'] = output_formats
            ext_list = config['input_extension']
            if not isinstance(ext_list, list):
                ext_list = [ext_list]
            for ext in ext_list:
                self.executor_model.add_input_executor_configuration(
                    ext, context_id, executor_id, config)
                executor_count += 1

                # Save default configs to our config system so that they are
                # displayed in the Run confpage
                config_widget = config["configuration_widget"]
                default_conf = (
                    config_widget.get_default_configuration()
                    if config_widget
                    else {}
                )
                self._save_default_graphical_executor_configuration(
                    executor_id,
                    ext,
                    context_id,
                    default_conf
                )

        self.executor_use_count[executor_id] = executor_count
        self.executor_model.set_executor_name(executor_id, executor_name)
        self.set_actions_status()

    def deregister_executor_configuration(
        self,
        executor: RunExecutor,
        configuration: List[SupportedExecutionRunConfiguration]
    ):
        """
        Deregister a :class:`RunExecutor` instance from providing a set
        of run configurations that are no longer supported by it.

        Parameters
        ----------
        executor: RunExecutor
            A :class:`SpyderPluginV2` instance that implements the
            :class:`RunExecutor` interface and will deregister execution
            input type information.
        configuration: List[SupportedExecutionRunConfiguration]
            A list of input configurations that the executor wants to
            deregister. Each configuration specifies the input extension
            identifier as well as the available execution context for that
            type.
        """
        executor_id = executor.NAME

        for config in configuration:
            ext = config['input_extension']
            context = config['context']
            context_name = context['name']
            context_id = getattr(RunContext, context_name)
            self.executor_model.remove_input_executor_configuration(
                ext, context_id, executor_id)
            self.executor_use_count[executor_id] -= 1

        if self.executor_use_count[executor_id] <= 0:
            self.run_executors.pop(executor_id)

        self.set_actions_status()

    def register_viewer_configuration(
        self,
        viewer: RunResultViewer,
        formats: List[OutputFormat]
    ):
        """
        Register a :class:`RunResultViewer` instance to indicate its support
        for a given set of output run result formats.

        Parameters
        ----------
        provider: RunResultViewer
            A :class:`SpyderPluginV2` instance that implements the
            :class:`RunResultViewer` interface and will register
            supported output formats.
        formats: List[OutputFormat]
            A list of output formats that the viewer is able to display.
        """
        for out_format in formats:
            format_name = out_format['name']
            format_id = out_format.get('identifier', None)
            if format_id is None:
                format_id = camel_case_to_snake_case(format_name)
            setattr(RunResultFormat, format_name, format_id)

            viewers_set = self.viewers.get(format_id, WeakSet())
            viewers_set.add(viewer)
            self.viewers[format_id] = viewers_set

    def deregister_viewer_configuration(
        self,
        viewer: RunResultViewer,
        formats: List[OutputFormat]
    ):
        """
        Deregister a :class:`RunResultViewer` instance from supporting a set of
        output formats that are no longer supported by it.

        Parameters
        ----------
        provider: RunResultViewer
            A :class:`SpyderPluginV2` instance that implements the
            :class:`RunResultViewer` interface and will deregister
            output format support.
        formats: List[OutputFormat]
            A list of output formats that the viewer wants to deregister.
        """
        for out_format in formats:
            format_name = out_format['name']
            format_id = getattr(RunResultFormat, format_name)
            if format_id in self.viewers:
                viewer_set = self.viewers[format_id]
                viewer_set.discard(viewer)

    def get_executor_configuration_parameters(
        self,
        executor_name: str,
        extension: str,
        context_id: str
    ) -> StoredRunExecutorParameters:
        """
        Retrieve the stored parameters for a given executor `executor_name`
        using context `context_id` with file extension `extension`.

        Parameters
        ----------
        executor_name: str
            The identifier of the run executor.
        extension: str
            The file extension to register the configuration parameters for.
        context_id: str
            The context to register the configuration parameters for.

        Returns
        -------
        config: StoredRunExecutorParameters
            A dictionary containing the run executor parameters for the given
            run configuration.
        """

        all_execution_params: Dict[
            str,
            Dict[Tuple[str, str], StoredRunExecutorParameters]
        ] = self.get_conf('parameters', default={})

        executor_params = all_execution_params.get(executor_name, {})
        params = executor_params.get(
            (extension, context_id),
            StoredRunExecutorParameters(params={})
        )

        return params

    def set_executor_configuration_parameters(
        self,
        executor_name: str,
        extension: str,
        context_id: str,
        params: StoredRunExecutorParameters
    ):
        """
        Update and save the list of configuration parameters for a given
        executor on a given pair of context and file extension.

        Parameters
        ----------
        executor_name: str
            The identifier of the run executor.
        extension: str
            The file extension to register the configuration parameters for.
        context_id: str
            The context to register the configuration parameters for.
        params: StoredRunExecutorParameters
            A dictionary containing the run configuration parameters for the
            given executor.
        """
        all_execution_params: Dict[
            str,
            Dict[Tuple[str, str], StoredRunExecutorParameters]
        ] = self.get_conf('parameters', default={})

        executor_params = all_execution_params.get(executor_name, {})
        ext_ctx_params = executor_params.get((extension, context_id), {})

        if ext_ctx_params:
            # Update current parameters in case the user has already saved some
            # before.
            ext_ctx_params['params'].update(params['params'])
        else:
            # Create a new entry of executor parameters in case there isn't any
            executor_params[(extension, context_id)] = params

        all_execution_params[executor_name] = executor_params

        self.set_conf('parameters', all_execution_params)

    def delete_executor_configuration_parameters(
        self,
        executor_name: str,
        extension: str,
        context_id: str,
        uuid: str
    ):
        """
        Delete an executor parameter set from our config system.

        Parameters
        ----------
        executor_name: str
            The identifier of the run executor.
        extension: str
            The file extension of the configuration parameters to delete.
        context_id: str
            The context of the configuration parameters to delete.
        uuid: str
            The run configuration identifier.
        """
        all_execution_params: Dict[
            str,
            Dict[Tuple[str, str], StoredRunExecutorParameters]
        ] = self.get_conf('parameters', default={})

        executor_params = all_execution_params[executor_name]
        ext_ctx_params = executor_params[(extension, context_id)]['params']

        for params_id in ext_ctx_params:
            if params_id == uuid:
                # Prevent to remove default parameters
                if ext_ctx_params[params_id]["default"]:
                    return

                ext_ctx_params.pop(params_id, None)
                break

        executor_params[(extension, context_id)]['params'] = ext_ctx_params
        all_execution_params[executor_name] = executor_params

        self.set_conf('parameters', all_execution_params)

    def get_last_used_executor_parameters(
        self,
        uuid: str
    ) -> StoredRunConfigurationExecutor:
        """
        Retrieve the last used execution parameters for a given
        run configuration.

        Parameters
        ----------
        uuid: str
            The run configuration identifier.

        Returns
        -------
        last_used_params: StoredRunConfigurationExecutor
            A dictionary containing the last used executor and parameters
            for the given run configuration.
        """
        last_executor = self._get_last_configured_executor(uuid)
        default = StoredRunConfigurationExecutor(executor=None, selected=None)

        if last_executor is None:
            return default
        else:
            mru_executors_uuids: Dict[
                str, List[StoredRunConfigurationExecutor]
            ] = self.get_conf('last_used_parameters_per_executor', default={})

            all_params = mru_executors_uuids.get(uuid, [])

            last_used_params = default
            for params in all_params:
                if params["executor"] == last_executor:
                    last_used_params = params

            return last_used_params

    def get_last_used_execution_params(
        self,
        uuid: str,
        executor_name: str
    ) -> Optional[str]:
        """
        Retrieve the last used execution parameters for a given pair of run
        configuration identifier and executor name.

        Parameters
        ----------
        uuid: str
            The run configuration identifier.
        executor_name: str
            The identifier of the run executor.

        Returns
        -------
        last_used_params: Optional[str]
            The identifier of the last used parameters for the given
            run configuration on the given executor. None if the executor has
            not executed the run configuration.
        """

        mru_executors_uuids: Dict[
            str,
            List[StoredRunConfigurationExecutor]
        ] = self.get_conf('last_used_parameters_per_executor', default={})

        default = StoredRunConfigurationExecutor(
            executor=executor_name,
            selected=None
        )
        all_params = mru_executors_uuids.get(uuid, [default])

        last_used_params = None
        for params in all_params:
            if params['executor'] == executor_name:
                last_used_params = params['selected']

        return last_used_params

    def set_last_used_execution_params(
        self,
        uuid: str,
        params: StoredRunConfigurationExecutor
    ):
        """
        Store the list of last used executors and their parameters for a given
        run configuration.

        Parameters
        ----------
        uuid: str
            The run configuration identifier.
        params: StoredRunConfigurationExecutor
            Dictionary containing the last used executor and run parameters
            used.
        """
        mru_executors_uuids: Dict[
            str,
            List[StoredRunConfigurationExecutor]
        ] = self.get_conf('last_used_parameters_per_executor', default={})

        if uuid in mru_executors_uuids:
            new_params = []
            params_added = False
            for saved_params in mru_executors_uuids[uuid]:
                # Discard saved params for the executor and use the new ones
                # instead. We need to do this when the executor was configured
                # before
                if params["executor"] == saved_params["executor"]:
                    new_params.append(params)
                    params_added = True
                else:
                    # Add params for all other executors
                    new_params.append(saved_params)

            # If the executor hasn't been configured yet, we simply need to add
            # its new params to the list.
            if not params_added:
                new_params.append(params)

            # Replace current params with the new ones
            mru_executors_uuids[uuid] = new_params
        else:
            mru_executors_uuids[uuid] = [params]

        self.set_conf('last_used_parameters_per_executor', mru_executors_uuids)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _save_default_graphical_executor_configuration(
        self,
        executor_name: str,
        extension: str,
        context_id: str,
        default_conf: dict,
    ):
        """
        Save a default executor configuration to our config system.

        Parameters
        ----------
        executor_name: str
            The identifier of the run executor.
        extension: str
            The file extension to register the configuration parameters for.
        context_id: str
            The context to register the configuration parameters for.
        default_conf: dict
            A dictionary containing the run configuration parameters for the
            given executor.
        """
        # Check if there's already a default parameter config to not do this
        # because it's not necessary.
        current_params = self.get_executor_configuration_parameters(
            executor_name,
            extension,
            context_id
        )

        for param in current_params["params"].values():
            if param["default"]:
                return

        # Id for this config
        uuid = str(uuid4())

        # Build config
        cwd_opts = WorkingDirOpts(
            source=WorkingDirSource.ConfigurationDirectory,
            path=None
        )

        exec_params = RunExecutionParameters(
            working_dir=cwd_opts, executor_params=default_conf
        )

        ext_exec_params = ExtendedRunExecutionParameters(
            uuid=uuid,
            name=_("Default"),
            params=exec_params,
            file_uuid=None,
            default=True,
        )

        store_params = StoredRunExecutorParameters(
            params={uuid: ext_exec_params}
        )

        # Save config
        self.set_executor_configuration_parameters(
            executor_name,
            extension,
            context_id,
            store_params,
        )

    def _get_last_configured_executor(self, uuid: str) -> Optional[str]:
        """
        Get the last executor configured in RunDialog for a given file.

        Parameters
        ----------
        uuid: str
            Unique id associated to a file.
        """
        mru_uuids_to_executor_names: Dict[
            str, str,
        ] = self.get_conf('last_configured_executor', default={})

        return mru_uuids_to_executor_names.get(uuid, None)

    def _set_last_configured_executor(self, uuid: str, executor_name: str):
        """
        Set the last executor configured in RunDialog for a given file.

        Parameters
        ----------
        uuid: str
            Unique id associated to a file.
        """
        mru_uuids_to_executor_names: Dict[
            str, str,
        ] = self.get_conf('last_configured_executor', default={})

        mru_uuids_to_executor_names[uuid] = executor_name
        self.set_conf('last_configured_executor', mru_uuids_to_executor_names)

    def _get_default_executor(self, extension: str) -> Optional[str]:
        """
        Get the default for a given file extension.

        Parameters
        ----------
        extension: str
            The file extension

        Returns
        -------
        executor_name: Optional[str]
            The default executor name or None if there isn't one for the given
            extension.

        Notes
        -----
        - This is necessary in case a file has multiple executors so that the
          Run file, cell and selection buttons use this executor by default.
        - We need to implement an API and UI for this so that a default
          executor can be declared for other file types.
        """
        executor_name = None

        if extension in ["py", "ipy", "pyx"]:
            # This is the most intuitive executor for Python-like files and it
            # was also the default in Spyder 5.
            executor_name = Plugins.IPythonConsole

        # This avoids changing the last executor selected in RunDialog when
        # using the default one.
        self._save_last_configured_executor = False

        return executor_name



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Variable Explorer Main Plugin Widget.
"""

# Third party imports
from qtpy.QtCore import QTimer, Slot, Signal
from qtpy.QtWidgets import QAction

# Local imports
from spyder.api.config.decorators import on_conf_change
from spyder.api.translations import _
from spyder.api.shellconnect.main_widget import ShellConnectMainWidget
from spyder.plugins.variableexplorer.widgets.namespacebrowser import (
    NamespaceBrowser)
from spyder.utils.icon_manager import ima
from spyder.utils.programs import is_module_installed


# =============================================================================
# ---- Constants
# =============================================================================
class VariableExplorerWidgetActions:
    # Triggers
    ImportData = 'import_data_action'
    SaveData = 'save_data_action'
    SaveDataAs = 'save_data_as_action'
    ResetNamespace = 'reset_namespaces_action'
    Search = 'search'
    Refresh = 'refresh'

    # Toggles
    ToggleExcludePrivate = 'toggle_exclude_private_action'
    ToggleExcludeUpperCase = 'toggle_exclude_uppercase_action'
    ToggleExcludeCapitalized = 'toggle_exclude_capitalized_action'
    ToggleExcludeUnsupported = 'toggle_exclude_unsupported_action'
    ToggleExcludeCallablesAndModules = (
        'toggle_exclude_callables_and_modules_action')
    ToggleMinMax = 'toggle_minmax_action'
    ToggleFilter = 'toggle_filter_variable_action'

    # Resize
    ResizeRowsAction = 'resize_rows_action'
    ResizeColumnsAction = 'resize_columns_action'


class VariableExplorerWidgetOptionsMenuSections:
    Display = 'excludes_section'
    Highlight = 'highlight_section'
    Resize = 'resize_section'


class VariableExplorerWidgetMainToolBarSections:
    Main = 'main_section'


class VariableExplorerWidgetMenus:
    EmptyContextMenu = 'empty'
    PopulatedContextMenu = 'populated'


class VariableExplorerContextMenuActions:
    PasteAction = 'paste_action'
    CopyAction = 'copy'
    EditAction = 'edit_action'
    PlotAction = 'plot_action'
    HistogramAction = 'histogram_action'
    ImshowAction = 'imshow_action'
    SaveArrayAction = 'save_array_action'
    InsertAction = 'insert_action'
    RemoveAction = 'remove_action'
    RenameAction = 'rename_action'
    DuplicateAction = 'duplicate_action'
    ViewAction = 'view_action'
    EditFiltersAction = 'edit_filters_action'


class VariableExplorerContextMenuSections:
    Edit = 'edit_section'
    Insert = 'insert_section'
    View = 'view_section'
    Filter = 'Filter_section'


# =============================================================================
# ---- Widgets
# =============================================================================

class VariableExplorerWidget(ShellConnectMainWidget):

    # PluginMainWidget class constants
    ENABLE_SPINNER = True

    # Other class constants
    INITIAL_FREE_MEMORY_TIME_TRIGGER = 60 * 1000  # ms
    SECONDARY_FREE_MEMORY_TIME_TRIGGER = 180 * 1000  # ms

    sig_open_preferences_requested = Signal()
    """
    Signal to open the variable explorer preferences.
    """

    sig_show_figure_requested = Signal(bytes, str, object)
    """
    This is emitted to request that a figure be shown in the Plots plugin.

    Parameters
    ----------
    image: bytes
        The image to show.
    mime_type: str
        The image's mime type.
    shellwidget: ShellWidget
        The shellwidget associated with the figure.
    """

    def __init__(self, name=None, plugin=None, parent=None):
        super().__init__(name, plugin, parent)

        # Widgets
        self.context_menu = None
        self.empty_context_menu = None
        self.filter_button = None

        # Attributes
        self._is_filter_button_checked = True
        self.plots_plugin_enabled = False

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def get_title(self):
        return _('Variable Explorer')

    def setup(self):
        # ---- Options menu actions
        self.show_minmax_action = self.create_action(
            VariableExplorerWidgetActions.ToggleMinMax,
            text=_("Show arrays min/max"),
            tip=_("Show minimum and maximum of arrays"),
            toggled=True,
            option='minmax'
        )

        # ---- Toolbar actions
        import_data_action = self.create_action(
            VariableExplorerWidgetActions.ImportData,
            text=_('Import data'),
            icon=self.create_icon('fileimport'),
            triggered=lambda x: self.import_data(),
        )

        save_action = self.create_action(
            VariableExplorerWidgetActions.SaveData,
            text=_("Save data"),
            icon=self.create_icon('filesave'),
            triggered=lambda x: self.save_data(),
        )

        save_as_action = self.create_action(
            VariableExplorerWidgetActions.SaveDataAs,
            text=_("Save data as..."),
            icon=self.create_icon('filesaveas'),
            triggered=lambda x: self.save_data(),
        )

        reset_namespace_action = self.create_action(
            VariableExplorerWidgetActions.ResetNamespace,
            text=_("Remove all variables"),
            icon=self.create_icon('editdelete'),
            triggered=lambda x: self.reset_namespace(),
        )

        # ---- Context menu actions
        resize_rows_action = self.create_action(
            VariableExplorerWidgetActions.ResizeRowsAction,
            text=_("Resize rows to contents"),
            icon=self.create_icon('collapse_row'),
            triggered=self.resize_rows
        )

        resize_columns_action = self.create_action(
            VariableExplorerWidgetActions.ResizeColumnsAction,
            _("Resize columns to contents"),
            icon=self.create_icon('collapse_column'),
            triggered=self.resize_columns
        )

        self.paste_action = self.create_action(
            VariableExplorerContextMenuActions.PasteAction,
            _("Paste"),
            icon=self.create_icon('editpaste'),
            triggered=self.paste
        )

        self.copy_action = self.create_action(
            VariableExplorerContextMenuActions.CopyAction,
            _("Copy"),
            icon=self.create_icon('editcopy'),
            triggered=self.copy
        )

        self.edit_action = self.create_action(
            VariableExplorerContextMenuActions.EditAction,
            _("Edit"),
            icon=self.create_icon('edit'),
            triggered=self.edit_item
        )

        self.plot_action = self.create_action(
            VariableExplorerContextMenuActions.PlotAction,
            _("Plot"),
            icon=self.create_icon('plot'),
            triggered=self.plot_item
        )
        self.plot_action.setVisible(False)

        self.hist_action = self.create_action(
            VariableExplorerContextMenuActions.HistogramAction,
            _("Histogram"),
            icon=self.create_icon('hist'),
            triggered=self.histogram_item
        )
        self.hist_action.setVisible(False)

        self.imshow_action = self.create_action(
            VariableExplorerContextMenuActions.ImshowAction,
            _("Show image"),
            icon=self.create_icon('imshow'),
            triggered=self.imshow_item
        )
        self.imshow_action.setVisible(False)

        self.save_array_action = self.create_action(
            VariableExplorerContextMenuActions.SaveArrayAction,
            _("Save array"),
            icon=self.create_icon('filesave'),
            triggered=self.save_array
        )
        self.save_array_action.setVisible(False)

        self.insert_action = self.create_action(
            VariableExplorerContextMenuActions.InsertAction,
            _("Insert"),
            icon=self.create_icon('insert'),
            triggered=self.insert_item
        )

        self.edit_filters = self.create_action(
            VariableExplorerContextMenuActions.EditFiltersAction,
            _("Edit filters"),
            icon=self.create_icon('filter'),
            triggered=self.sig_open_preferences_requested
        )

        self.remove_action = self.create_action(
            VariableExplorerContextMenuActions.RemoveAction,
            _("Remove"),
            icon=self.create_icon('editdelete'),
            triggered=self.remove_item
        )

        self.rename_action = self.create_action(
            VariableExplorerContextMenuActions.RenameAction,
            _("Rename"),
            icon=self.create_icon('rename'),
            triggered=self.rename_item
        )

        self.duplicate_action = self.create_action(
            VariableExplorerContextMenuActions.DuplicateAction,
            _("Duplicate"),
            icon=self.create_icon('edit_add'),
            triggered=self.duplicate_item
        )

        self.view_action = self.create_action(
            VariableExplorerContextMenuActions.ViewAction,
            _("View with the Object Explorer"),
            icon=self.create_icon('outline_explorer'),
            triggered=self.view_item
        )

        # Options menu
        options_menu = self.get_options_menu()
        for item in [self.exclude_private_action,
                     self.exclude_uppercase_action,
                     self.exclude_capitalized_action,
                     self.exclude_unsupported_action,
                     self.exclude_callables_and_modules_action,
                     self.show_minmax_action]:
            self.add_item_to_menu(
                item,
                menu=options_menu,
                section=VariableExplorerWidgetOptionsMenuSections.Display,
            )

        self._enable_filter_actions(self.get_conf('filter_on'))

        # Resize
        for item in [resize_rows_action, resize_columns_action]:
            self.add_item_to_menu(
                item,
                menu=options_menu,
                section=VariableExplorerWidgetOptionsMenuSections.Resize,
            )

        # Main toolbar
        main_toolbar = self.get_main_toolbar()
        for item in [import_data_action, save_action, save_as_action,
                     reset_namespace_action]:
            self.add_item_to_toolbar(
                item,
                toolbar=main_toolbar,
                section=VariableExplorerWidgetMainToolBarSections.Main,
            )
        save_action.setEnabled(False)

        # Search, Filter and Refresh buttons are added in _setup()

        # ---- Context menu to show when there are variables present
        self.context_menu = self.create_menu(
            VariableExplorerWidgetMenus.PopulatedContextMenu)
        for item in [self.edit_action, self.copy_action, self.paste_action,
                     self.rename_action, self.remove_action,
                     self.save_array_action]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=VariableExplorerContextMenuSections.Edit,
            )

        for item in [self.insert_action, self.duplicate_action]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=VariableExplorerContextMenuSections.Insert,
            )

        for item in [self.edit_filters]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=VariableExplorerContextMenuSections.Filter,
            )

        for item in [self.view_action, self.plot_action, self.hist_action,
                     self.imshow_action]:
            self.add_item_to_menu(
                item,
                menu=self.context_menu,
                section=VariableExplorerContextMenuSections.View,
            )

        # ---- Context menu when the variable explorer is empty
        self.empty_context_menu = self.create_menu(
            VariableExplorerWidgetMenus.EmptyContextMenu)
        for item in [self.insert_action, self.paste_action]:
            self.add_item_to_menu(
                item,
                menu=self.empty_context_menu,
                section=VariableExplorerContextMenuSections.Edit,
            )

    def _setup(self):
        """
        Create options menu and adjacent toolbar buttons, etc.

        This creates base actions related with Search, Filter and Refresh.

        This calls the parent's method to setup default actions, create the
        spinner and the options menu, and connect signals. After that, it adds
        the Search, Filter and Refresh buttons between the spinner and the
        options menu.
        """
        super()._setup()

        # ---- Base Options menu actions
        self.exclude_private_action = self.create_action(
            VariableExplorerWidgetActions.ToggleExcludePrivate,
            text=_("Exclude private variables"),
            tip=_("Exclude variables that start with an underscore"),
            toggled=True,
            option='exclude_private',
        )

        self.exclude_uppercase_action = self.create_action(
            VariableExplorerWidgetActions.ToggleExcludeUpperCase,
            text=_("Exclude all-uppercase variables"),
            tip=_("Exclude variables whose name is uppercase"),
            toggled=True,
            option='exclude_uppercase',
        )

        self.exclude_capitalized_action = self.create_action(
            VariableExplorerWidgetActions.ToggleExcludeCapitalized,
            text=_("Exclude capitalized variables"),
            tip=_("Exclude variables whose name starts with a capital "
                  "letter"),
            toggled=True,
            option='exclude_capitalized',
        )

        self.exclude_unsupported_action = self.create_action(
            VariableExplorerWidgetActions.ToggleExcludeUnsupported,
            text=_("Exclude unsupported data types"),
            tip=_("Exclude references to data types that don't have "
                  "an specialized viewer or can't be edited."),
            toggled=True,
            option='exclude_unsupported',
        )

        self.exclude_callables_and_modules_action = self.create_action(
            VariableExplorerWidgetActions.ToggleExcludeCallablesAndModules,
            text=_("Exclude callables and modules"),
            tip=_("Exclude references to functions, modules and "
                  "any other callable."),
            toggled=True,
            option='exclude_callables_and_modules'
        )

        # ---- Base Toolbar actions
        self.search_action = self.create_action(
            VariableExplorerWidgetActions.Search,
            text=_("Search variable names and types"),
            icon=self.create_icon('find'),
            toggled=self.toggle_finder,
            register_shortcut=True
        )

        self.refresh_action = self.create_action(
            VariableExplorerWidgetActions.Refresh,
            text=_("Refresh variables"),
            icon=self.create_icon('refresh'),
            triggered=self.refresh_table,
            register_shortcut=True,
        )

        self.filter_button = self.create_action(
            VariableExplorerWidgetActions.ToggleFilter,
            text="",
            icon=ima.icon('filter'),
            toggled=self._enable_filter_actions,
            option='filter_on',
            tip=_("Filter variables")
        )
        self.filter_button.setCheckable(True)
        self.filter_button.toggled.connect(self._set_filter_button_state)

        for action in [
            self.search_action,
            self.filter_button,
            self.refresh_action,
        ]:
            self.add_corner_widget(action, before=self._options_button)

    def update_actions(self):
        """Update the actions."""
        if self.is_current_widget_empty():
            self._set_main_toolbar_state(False)
            return
        else:
            self._set_main_toolbar_state(True)

        action = self.get_action(VariableExplorerWidgetActions.ToggleMinMax)
        action.setEnabled(is_module_installed('numpy'))

        nsb = self.current_widget()
        if nsb:
            save_data_action = self.get_action(
                VariableExplorerWidgetActions.SaveData)
            save_data_action.setEnabled(nsb.filename is not None)

        search_action = self.get_action(VariableExplorerWidgetActions.Search)
        if nsb is None:
            checked = False
        else:
            checked = nsb.finder_is_visible()
        search_action.setChecked(checked)

    @on_conf_change
    def on_section_conf_change(self, section):
        for index in range(self.count()):
            widget = self._stack.widget(index)
            if widget:
                widget.setup()

    def set_plots_plugin_enabled(self, value: bool):
        """
        Change whether the Plots plugin is enabled.

        This stores the information in this widget and propagates it to every
        NamespaceBrowser.
        """
        self.plots_plugin_enabled = value
        for index in range(self.count()):
            nsb = self._stack.widget(index)
            if nsb:
                nsb.plots_plugin_enabled = value

    # ---- Stack accesors
    # ------------------------------------------------------------------------
    def switch_widget(self, nsb, old_nsb):
        """Set the current NamespaceBrowser."""
        pass

    # ---- Public API
    # ------------------------------------------------------------------------
    def create_new_widget(self, shellwidget):
        """Create new NamespaceBrowser."""
        nsb = NamespaceBrowser(self)
        nsb.sig_hide_finder_requested.connect(self.hide_finder)
        nsb.sig_free_memory_requested.connect(self.free_memory)
        nsb.sig_start_spinner_requested.connect(self.start_spinner)
        nsb.sig_stop_spinner_requested.connect(self.stop_spinner)
        nsb.sig_show_figure_requested.connect(self.sig_show_figure_requested)
        nsb.set_shellwidget(shellwidget)
        nsb.plots_plugin_enabled = self.plots_plugin_enabled
        nsb.setup()
        self._set_actions_and_menus(nsb)

        # To update the Variable Explorer after execution
        shellwidget.sig_kernel_state_arrived.connect(nsb.update_view)
        shellwidget.sig_config_spyder_kernel.connect(
            nsb.set_namespace_view_settings
        )
        return nsb

    def close_widget(self, nsb):
        """Close NamespaceBrowser."""
        nsb.sig_hide_finder_requested.disconnect(self.hide_finder)
        nsb.sig_free_memory_requested.disconnect(self.free_memory)
        nsb.sig_start_spinner_requested.disconnect(self.start_spinner)
        nsb.sig_stop_spinner_requested.disconnect(self.stop_spinner)
        nsb.sig_show_figure_requested.disconnect(
            self.sig_show_figure_requested)
        nsb.shellwidget.sig_kernel_state_arrived.disconnect(nsb.update_view)
        nsb.shellwidget.sig_config_spyder_kernel.disconnect(
            nsb.set_namespace_view_settings
        )

        nsb.close()
        nsb.setParent(None)

    def import_data(self, filenames=None):
        """
        Import data in current namespace.
        """
        if not self.is_current_widget_empty():
            nsb = self.current_widget()
            nsb.refresh_table()
            nsb.import_data(filenames=filenames)

    def save_data(self):
        if not self.is_current_widget_empty():
            nsb = self.current_widget()
            nsb.save_data()
            self.update_actions()

    def reset_namespace(self):
        if not self.is_current_widget_empty():
            nsb = self.current_widget()
            nsb.reset_namespace()

    @Slot(bool)
    def toggle_finder(self, checked):
        """Hide or show the finder."""
        widget = self.current_widget()
        if widget is None or self.is_current_widget_empty():
            return
        widget.toggle_finder(checked)

    @Slot()
    def hide_finder(self):
        """Hide the finder."""
        action = self.get_action(VariableExplorerWidgetActions.Search)
        action.setChecked(False)

    def refresh_table(self):
        if not self.is_current_widget_empty():
            nsb = self.current_widget()
            nsb.refresh_table()

    @Slot()
    def free_memory(self):
        """
        Free memory signal.
        """
        self.sig_free_memory_requested.emit()
        QTimer.singleShot(self.INITIAL_FREE_MEMORY_TIME_TRIGGER,
                          self.sig_free_memory_requested)
        QTimer.singleShot(self.SECONDARY_FREE_MEMORY_TIME_TRIGGER,
                          self.sig_free_memory_requested)

    def resize_rows(self):
        if self._current_editor is not None:
            self._current_editor.resizeRowsToContents()

    def resize_columns(self):
        if self._current_editor is not None:
            self._current_editor.resize_column_contents()

    def paste(self):
        self._current_editor.paste()

    def copy(self):
        self._current_editor.copy()

    def edit_item(self):
        self._current_editor.edit_item()

    def plot_item(self):
        self._current_editor.plot_item('plot')

    def histogram_item(self):
        self._current_editor.plot_item('hist')

    def imshow_item(self):
        self._current_editor.imshow_item()

    def save_array(self):
        self._current_editor.save_array()

    def insert_item(self):
        self._current_editor.insert_item(below=False)

    def remove_item(self):
        self._current_editor.remove_item()

    def rename_item(self):
        self._current_editor.rename_item()

    def duplicate_item(self):
        self._current_editor.duplicate_item()

    def view_item(self):
        self._current_editor.view_item()

    # ---- Private API
    # ------------------------------------------------------------------------
    @property
    def _current_editor(self):
        editor = None
        if not self.is_current_widget_empty():
            nsb = self.current_widget()
            editor = nsb.editor
        return editor

    def _set_actions_and_menus(self, nsb):
        """
        Set actions and menus created here and used by the namespace
        browser editor.

        Although this is not ideal, it's necessary to be able to use
        the CollectionsEditor widget separately from this plugin.
        """
        editor = nsb.editor

        # Actions
        editor.paste_action = self.paste_action
        editor.copy_action = self.copy_action
        editor.edit_action = self.edit_action
        editor.plot_action = self.plot_action
        editor.hist_action = self.hist_action
        editor.imshow_action = self.imshow_action
        editor.save_array_action = self.save_array_action
        editor.insert_action = self.insert_action
        editor.remove_action = self.remove_action
        editor.rename_action = self.rename_action
        editor.duplicate_action = self.duplicate_action
        editor.view_action = self.view_action

        # Menus
        editor.menu = self.context_menu
        editor.empty_ws_menu = self.empty_context_menu

        # These actions are not used for dictionaries (so we don't need them
        # for namespaces) but we have to create them so they can be used in
        # several places in CollectionsEditor.
        editor.insert_action_above = QAction()
        editor.insert_action_below = QAction()

    def _enable_filter_actions(self, value):
        """Handle the change of the filter state."""
        self.exclude_private_action.setEnabled(value)
        self.exclude_uppercase_action.setEnabled(value)
        self.exclude_capitalized_action.setEnabled(value)
        self.exclude_unsupported_action.setEnabled(value)
        self.exclude_callables_and_modules_action.setEnabled(value)

    def _set_main_toolbar_state(self, enabled):
        """Set main toolbar enabled state."""
        main_toolbar = self.get_main_toolbar()
        for action in main_toolbar.actions():
            action.setEnabled(enabled)

        # Adjustments for the filter button
        if enabled:
            # Restore state for active consoles
            self.filter_button.setChecked(self._is_filter_button_checked)
        else:
            # Uncheck button for dead consoles if it's checked so that the
            # toolbar looks good
            if self.filter_button.isChecked():
                self.filter_button.setChecked(False)
                self._is_filter_button_checked = True

    def _set_filter_button_state(self, checked):
        """Keep track of the filter button checked state."""
        self._is_filter_button_checked = checked



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Conda/anaconda utilities."""

# Standard library imports
from functools import lru_cache
from glob import glob
import json
import os
import os.path as osp
import sys

# Third-party imports
from packaging.version import parse
from spyder_kernels.utils.pythonenv import (
    add_quotes,
    get_conda_env_path,
    is_conda_env,
)

# Local imports
from spyder.utils.programs import find_program, run_program, run_shell_command
from spyder.config.base import is_conda_based_app

WINDOWS = os.name == 'nt'
CONDA_ENV_LIST_CACHE = {}


def _env_for_conda():
    """
    Required environment variables for Conda to work as expected.

    Notes
    -----
    This is needed on Windows since Conda 23.9.0
    """
    env = {}
    if os.name == 'nt':
        env_vars = [("HOMEDRIVE", "C:"), ("HOMEPATH", "\\Users\\xxxxx")]
        for var, default in env_vars:
            value = os.environ.get(var, default)
            env[var] = value

    return env


def get_conda_root_prefix(pyexec=None, quote=False):
    """
    Return conda prefix from pyexec path

    If `quote` is True, then quotes are added if spaces are found in the path.
    """
    if pyexec is None:
        conda_env_prefix = sys.prefix
    else:
        conda_env_prefix = get_conda_env_path(pyexec)

    conda_env_prefix = conda_env_prefix.replace('\\', '/')
    env_key = '/envs/'

    if conda_env_prefix.rfind(env_key) != -1:
        root_prefix = conda_env_prefix.split(env_key)[0]
    else:
        root_prefix = conda_env_prefix

    if quote:
        root_prefix = add_quotes(root_prefix)

    return root_prefix


def find_conda(pyexec=None):
    """
    Find conda executable.

    `pyexec` is a python executable, the relative location from which to
    attempt to locate a conda executable.
    """
    conda = None

    # First try Spyder's conda executable
    if is_conda_based_app():
        root = osp.dirname(os.environ['CONDA_EXE'])
        conda = osp.join(root, 'conda.exe' if WINDOWS else 'conda')

    # Next try the environment variables
    if conda is None:
        conda = os.environ.get('CONDA_EXE') or os.environ.get('MAMBA_EXE')

    # Next try searching for the executable
    if conda is None:
        conda_exec = 'conda.bat' if WINDOWS else 'conda'
        extra_paths = [
            osp.join(get_conda_root_prefix(_pyexec), 'condabin')
            for _pyexec in [sys.executable, pyexec]
        ]
        conda = find_program(conda_exec, extra_paths)

    return conda


def get_list_conda_envs():
    """Return the list of all conda envs found in the system."""
    global CONDA_ENV_LIST_CACHE

    env_list = {}
    conda = find_conda()
    if conda is None:
        return env_list

    cmdstr = ' '.join([conda, 'env', 'list', '--json'])

    try:
        out, __ = run_shell_command(cmdstr, env=_env_for_conda()).communicate()
        out = out.decode()
        out = json.loads(out)
    except Exception:
        out = {'envs': []}

    for env in out['envs']:
        data = env.split(osp.sep)
        name = data[-1]
        path = osp.join(env, 'python.exe') if WINDOWS else osp.join(
            env, 'bin', 'python')

        if (
            # In case the environment doesn't have Python
            not osp.isfile(path)
            # Don't list the installers base env
            or (is_conda_based_app(pyexec=path) and name != "spyder-runtime")
        ):
            continue

        try:
            version, __ = run_program(path, ['--version']).communicate()
            version = version.decode()
        except Exception:
            version = ''

        name = ('base' if name.lower().startswith('anaconda') or
                name.lower().startswith('miniconda') else name)
        name = 'Conda: {}'.format(name)

        if name in env_list:
            if not (path, version.strip()) == env_list[name]:
                prev_info = env_list[name]
                prev_data = prev_info[0]
                prev_data = prev_data.split(osp.sep)
                env_list.pop(name)
                index_common_folder = 1
                end_path = 1
                if not WINDOWS:
                    end_path = 2
                for i in range(-1, -len(data) - 1, -1):
                    if data[i] == prev_data[i - end_path]:
                        index_common_folder += 1
                    else:
                        break
                path_part = prev_data[
                    -index_common_folder - end_path : -end_path
                ]
                prev_name = (
                    f'Conda: '
                    f'{"/".join(path_part)}'
                )
                env_list[prev_name] = prev_info
                name = f'Conda: {"/".join(data[-index_common_folder:])}'

        env_list[name] = (path, version.strip())

    CONDA_ENV_LIST_CACHE = env_list
    return env_list


def get_list_conda_envs_cache():
    """Return a cache of envs to avoid computing them again."""
    return CONDA_ENV_LIST_CACHE


def is_anaconda_pkg(prefix=sys.prefix):
    """Detect if the anaconda meta package is installed."""
    if is_conda_env(prefix):
        conda_meta = osp.join(prefix, "conda-meta")
        if glob(f"{conda_meta}{os.sep}anaconda-[0-9]*.json"):
            return True

    return False


def get_spyder_conda_channel():
    """Get the conda channel from which Spyder was installed."""
    conda = find_conda()

    if conda is None:
        return None, None

    env = get_conda_env_path(sys.executable)
    cmdstr = ' '.join([conda, 'list', 'spyder', '--json', '--prefix', env])

    try:
        out, __ = run_shell_command(cmdstr, env=_env_for_conda()).communicate()
        out = out.decode()
        out = json.loads(out)
    except Exception:
        return None, None

    # Avoids iterating over non-dict objects
    if 'error' in out:
        return None, None

    # These variables can be unassigned after the next for, so we need to give
    # them a default value at this point.
    # Fixes spyder-ide/spyder#22054
    channel, channel_url = None, None

    for package_info in out:
        if package_info["name"] == 'spyder':
            channel = package_info["channel"]
            channel_url = package_info["base_url"]

    if channel_url is not None and "<develop>" in channel_url:
        channel_url = None

    return channel, channel_url


@lru_cache(maxsize=1)
def conda_version(conda_executable=None):
    """Get the conda version if available."""
    version = parse('0')
    if not conda_executable:
        conda_executable = find_conda()
    if not conda_executable:
        return version
    try:
        version, __ = run_program(conda_executable, ['--version']).communicate()
        version = parse(version.decode().split()[-1].strip())
    except Exception:
        pass
    return version



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""Tests for docstring generation."""

# Third party imports
import pytest
from qtpy.QtCore import Qt
from qtpy.QtGui import QTextCursor

# Local imports
from spyder.config.manager import CONF
from spyder.utils.qthelpers import qapplication
from spyder.plugins.editor.widgets.codeeditor import CodeEditor
from spyder.plugins.editor.extensions.docstring import FunctionInfo


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def editor_auto_docstring():
    """Set up Editor with auto docstring activated."""
    app = qapplication()
    editor = CodeEditor(parent=None)
    kwargs = {}
    kwargs['language'] = 'Python'
    kwargs['close_quotes'] = True
    kwargs['close_parentheses'] = True
    editor.setup_editor(**kwargs)
    return editor


# =============================================================================
# ---- Tests
# =============================================================================
@pytest.mark.parametrize(
    "text, indent, name_list, type_list, value_list, rtype",
    [
        ('def foo():', '', [], [], [], None),
        (""" def foo(arg0, arg1=':', arg2: str='-> (float, str):') -> \
             (float, int): """,
         ' ', ['arg0', 'arg1', 'arg2'], [None, None, 'str'],
         [None, "':'", "'-> (float, str):'"],
         '(float, int)')
    ])
def test_parse_function_definition(text, indent, name_list, type_list,
                                   value_list, rtype):
    """Test the parse_def method of FunctionInfo class."""
    func_info = FunctionInfo()
    func_info.parse_def(text)

    assert func_info.func_indent == indent
    assert func_info.arg_name_list == name_list
    assert func_info.arg_type_list == type_list
    assert func_info.arg_value_list == value_list
    assert func_info.return_type_annotated == rtype


@pytest.mark.parametrize(
    "text, indent, expected",
    [
        ("""    def foo():\n
        if 1:
            raise ValueError
        else:
            return\n
    class F:""",
         "    ",
         """\n        if 1:
            raise ValueError
        else:
            return\n"""),
        ("""def foo():
    return""",
         "",
         """    return""")
    ])
def test_get_function_body(editor_auto_docstring, text, indent, expected):
    """Test get function body."""
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.setPosition(0, QTextCursor.MoveAnchor)
    cursor.movePosition(QTextCursor.NextBlock)
    editor.setTextCursor(cursor)

    writer = editor.writer_docstring
    result = writer.get_function_body(indent)

    assert expected == result


@pytest.mark.parametrize("use_shortcut", [True, False])
@pytest.mark.parametrize(
    "doc_type, text, expected",
    [
        ('Numpydoc',
         '',
         ''
         ),
        ('Numpydoc',
         'if 1:\n    ',
         'if 1:\n    '
         ),
        ('Numpydoc',
         '''async def foo():
    raise
    raise ValueError
    raise ValueError("test")
    raise TypeError("test")
    yield ''',
         '''async def foo():
    """\n    \n
    Raises
    ------
    ValueError
        DESCRIPTION.
    TypeError
        DESCRIPTION.\n
    Yields
    ------
    None.

    """
    raise
    raise ValueError
    raise ValueError("test")
    raise TypeError("test")
    yield '''
         ),
        ('Numpydoc',
         '''  def foo():
      print('{}' % foo_raise Value)
      foo_yield''',
         '''  def foo():
      """\n      \n
      Returns
      -------
      None.

      """
      print('{}' % foo_raise Value)
      foo_yield''',
         ),
        ('Numpydoc',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    ''',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    """\n    \n
    Parameters
    ----------
    arg : TYPE
        DESCRIPTION.
    arg0 : TYPE
        DESCRIPTION.
    arg1 : int
        DESCRIPTION.
    arg2 : List[Tuple[str, float]]
        DESCRIPTION.
    arg3 : TYPE, optional
        DESCRIPTION. The default is '-> (float, int):'.
    arg4 : TYPE, optional
        DESCRIPTION. The default is ':float, int['.
    arg5 : str, optional
        DESCRIPTION. The default is '""'.

    Returns
    -------
    (List[Tuple[str, float]], str, float)
        DESCRIPTION.

    """
    '''),
        ('Googledoc',
         '''async def foo():
    raise
    raise ValueError
    raise TypeError("test")
    yield value
    ''',
         '''async def foo():
    """\n    \n
    Raises:
        ValueError: DESCRIPTION.
        TypeError: DESCRIPTION.\n
    Yields:
        value (TYPE): DESCRIPTION.

    """
    raise
    raise ValueError
    raise TypeError("test")
    yield value
    '''
         ),
        ('Googledoc',
         '''  def foo():
      ''',
         '''  def foo():
      """\n      \n
      Returns:
          None.

      """
      ''',
         ),
        ('Googledoc',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    ''',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    """\n    \n
    Args:
        arg (TYPE): DESCRIPTION.
        arg0 (TYPE): DESCRIPTION.
        arg1 (int): DESCRIPTION.
        arg2 (List[Tuple[str, float]]): DESCRIPTION.
        arg3 (TYPE, optional): DESCRIPTION. Defaults to '-> (float, int):'.
        arg4 (TYPE, optional): DESCRIPTION. Defaults to ':float, int['.
        arg5 (str, optional): DESCRIPTION. Defaults to '""'.

    Returns:
        (List[Tuple[str, float]], str, float): DESCRIPTION.

    """
    '''),
        ('Sphinxdoc',
         '''async def foo():
    raise
    raise ValueError
    raise TypeError("test")
    yield value
    ''',
         '''async def foo():
    """\n    \n    :raises ValueError: DESCRIPTION
    :raises TypeError: DESCRIPTION
    :yield: DESCRIPTION
    :rtype: TYPE

    """
    raise
    raise ValueError
    raise TypeError("test")
    yield value
    '''
         ),
        ('Sphinxdoc',
         '''  def foo():
      ''',
         '''  def foo():
      """\n      \n      :return: DESCRIPTION
      :rtype: TYPE

      """
      ''',
         ),
        ('Sphinxdoc',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    ''',
         '''def foo(arg, arg0, arg1: int, arg2: List[Tuple[str, float]],
    arg3='-> (float, int):', arg4=':float, int[', arg5: str='""') -> \
  (List[Tuple[str, float]], str, float):
    """\n    \n    :param arg: DESCRIPTION
    :type arg: TYPE
    :param arg0: DESCRIPTION
    :type arg0: TYPE
    :param arg1: DESCRIPTION
    :type arg1: int
    :param arg2: DESCRIPTION
    :type arg2: List[Tuple[str, float]]
    :param arg3: DESCRIPTION, defaults to '-> (float, int):'
    :type arg3: TYPE, optional
    :param arg4: DESCRIPTION, defaults to ':float, int['
    :type arg4: TYPE, optional
    :param arg5: DESCRIPTION, defaults to '""'
    :type arg5: str, optional
    :return: DESCRIPTION
    :rtype: (List[Tuple[str, float]], str, float)

    """
    ''')
    ])
def test_editor_docstring_by_shortcut(editor_auto_docstring, doc_type,
                                      text, expected, use_shortcut):
    """Test auto docstring by shortcut."""
    CONF.set('editor', 'docstring_type', doc_type)
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.setPosition(0, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)
    writer = editor.writer_docstring

    if use_shortcut:
        writer.write_docstring_for_shortcut()
    else:
        pos = editor.cursorRect().bottomRight()
        pos = editor.mapToGlobal(pos)
        writer.line_number_cursor = editor.get_line_number_at(pos)
        writer.write_docstring_at_first_line_of_function()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def foo():
      ''',
         '''  def foo():
      """\n      \n
      Returns
      -------
      None.

      """
      ''',)
    ])
def test_editor_docstring_below_def_by_shortcut(qtbot, editor_auto_docstring,
                                                text, expected):
    """Test auto docstring below function definition by shortcut."""
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.NextBlock)
    cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)

    editor.writer_docstring.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected, key',
    [
        ('''def foo():
''',
         '''def foo():
    """\n    \n
    Returns
    -------
    None.

    """''',
         Qt.Key_Enter),
        ('''def foo():
''',
         '''def foo():
    """a''',
         Qt.Key_A)
    ])
def test_editor_docstring_delayed_popup(qtbot, editor_auto_docstring,
                                        text, expected, key):
    """Test auto docstring using delayed popup."""
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.NextBlock)
    cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)

    qtbot.keyPress(editor, Qt.Key_Space)
    qtbot.keyPress(editor, Qt.Key_Space)
    qtbot.keyPress(editor, Qt.Key_Space)
    qtbot.keyPress(editor, Qt.Key_Space)
    qtbot.keyPress(editor, Qt.Key_QuoteDbl)
    qtbot.keyPress(editor, Qt.Key_QuoteDbl)
    qtbot.keyPress(editor, Qt.Key_QuoteDbl)
    qtbot.wait(1000)
    qtbot.keyPress(editor.menu_docstring, key)
    qtbot.wait(1000)

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def foo():
      raise
      foo_raise()
      raisefoo()
      raise ValueError
      is_yield()
      raise ValueError('tt')
      yieldfoo()
      \traise TypeError('tt')
      _yield
      ''',
         '''  def foo():
      """\n      \n
      Raises
      ------
      ValueError
          DESCRIPTION.
      TypeError
          DESCRIPTION.\n
      Returns
      -------
      None.

      """
      raise
      foo_raise()
      raisefoo()
      raise ValueError
      is_yield()
      raise ValueError('tt')
      yieldfoo()
      \traise TypeError('tt')
      _yield
      ''',),
        ('''def foo():
    return None
    return "f, b", v1, v2, 3.0, .7, (,), {}, [ab], f(a), None, a.b, a+b, True
    return "f, b", v1, v3, 420, 5., (,), {}, [ab], f(a), None, a.b, a+b, False
    ''',
         '''def foo():
    """\n    \n
    Returns
    -------
    str
        DESCRIPTION.
    v1 : TYPE
        DESCRIPTION.
    TYPE
        DESCRIPTION.
    numeric
        DESCRIPTION.
    float
        DESCRIPTION.
    tuple
        DESCRIPTION.
    dict
        DESCRIPTION.
    list
        DESCRIPTION.
    TYPE
        DESCRIPTION.
    TYPE
        DESCRIPTION.
    TYPE
        DESCRIPTION.
    TYPE
        DESCRIPTION.
    bool
        DESCRIPTION.

    """
    return None
    return "f, b", v1, v2, 3.0, .7, (,), {}, [ab], f(a), None, a.b, a+b, True
    return "f, b", v1, v3, 420, 5., (,), {}, [ab], f(a), None, a.b, a+b, False
    '''),
        ('''def foo():
    return no, (ano, eo, dken)
    ''',
         '''def foo():
    """\n    \n
    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    return no, (ano, eo, dken)
    ''')
    ])
def test_editor_docstring_with_body_numpydoc(qtbot, editor_auto_docstring,
                                             text, expected):
    """Test auto docstring of numpydoc when the function body is complex."""
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.setPosition(0, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)
    writer = editor.writer_docstring

    writer.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def foo():
      raise
      foo_raise()
      raisefoo()
      raise ValueError
      is_yield()
      raise ValueError('tt')
      yieldfoo()
      \traise TypeError('tt')
      _yield
      ''',
         '''  def foo():
      """\n      \n
      Raises:
          ValueError: DESCRIPTION.
          TypeError: DESCRIPTION.\n
      Returns:
          None.

      """
      raise
      foo_raise()
      raisefoo()
      raise ValueError
      is_yield()
      raise ValueError('tt')
      yieldfoo()
      \traise TypeError('tt')
      _yield
      ''',),
        ('''def foo():
    return None
    return "f, b", v1, v2, 3.0, .7, (,), {}, [ab], f(a), None, a.b, a+b, True
    return "f, b", v1, v3, 420, 5., (,), {}, [ab], f(a), None, a.b, a+b, False
    ''',
         '''def foo():
    """\n    \n
    Returns:
        str: DESCRIPTION.
        v1 (TYPE): DESCRIPTION.
        TYPE: DESCRIPTION.
        numeric: DESCRIPTION.
        float: DESCRIPTION.
        tuple: DESCRIPTION.
        dict: DESCRIPTION.
        list: DESCRIPTION.
        TYPE: DESCRIPTION.
        TYPE: DESCRIPTION.
        TYPE: DESCRIPTION.
        TYPE: DESCRIPTION.
        bool: DESCRIPTION.

    """
    return None
    return "f, b", v1, v2, 3.0, .7, (,), {}, [ab], f(a), None, a.b, a+b, True
    return "f, b", v1, v3, 420, 5., (,), {}, [ab], f(a), None, a.b, a+b, False
    '''),
        ('''def foo():
    return no, (ano, eo, dken)
    ''',
         '''def foo():
    """\n    \n
    Returns:
        TYPE: DESCRIPTION.

    """
    return no, (ano, eo, dken)
    ''')
    ])
def test_editor_docstring_with_body_googledoc(qtbot, editor_auto_docstring,
                                              text, expected):
    """Test auto docstring of googledoc when the function body is complex."""
    CONF.set('editor', 'docstring_type', 'Googledoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.setPosition(0, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)
    writer = editor.writer_docstring

    writer.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def test(self) -> Annotated[str, int("2")]:
      ''',
         '''  def test(self) -> Annotated[str, int("2")]:
      """\n      \n
      Returns
      -------
      Annotated[str, int("2")]
          DESCRIPTION.

      """
      ''',)
    ])
def test_docstring_annotated_call(editor_auto_docstring, text, expected):
    """
    Test auto docstring with annotated function call.

    This is a regression tests for issue spyder-ide/spyder#14520
    """
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.NextBlock)
    cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)

    editor.writer_docstring.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def test(v:
           int):
      ''',
         '''  def test(v:
           int):
      """\n      \n
      Parameters
      ----------
      v : int
          DESCRIPTION.

      Returns
      -------
      None.

      """
      ''',)
    ])
def test_docstring_line_break(editor_auto_docstring, text, expected):
    """
    Test auto docstring with function call with line breaks.

    This is a regression tests for issue spyder-ide/spyder#14521
    """
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.NextBlock)
    cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)

    editor.writer_docstring.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ('''  def test(v: str = "#"):  # comment, with '#' and "#"
      ''',
         '''  def test(v: str = "#"):  # comment, with '#' and "#"
      """\n      \n
      Parameters
      ----------
      v : str, optional
          DESCRIPTION. The default is "#".

      Returns
      -------
      None.

      """
      '''),
        ('''  def test(v1: str = "#", # comment, with '#' and "#"
         v2: str = '#') -> str:
      ''',
         '''  def test(v1: str = "#", # comment, with '#' and "#"
         v2: str = '#') -> str:
      """\n      \n
      Parameters
      ----------
      v1 : str, optional
          DESCRIPTION. The default is "#".
      v2 : str, optional
          DESCRIPTION. The default is '#'.

      Returns
      -------
      str
          DESCRIPTION.

      """
      '''),
    ])
def test_docstring_comments(editor_auto_docstring, text, expected):
    """
    Test auto docstring with comments on lines of function definition.
    """
    CONF.set('editor', 'docstring_type', 'Numpydoc')
    editor = editor_auto_docstring
    editor.set_text(text)

    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.NextBlock)
    cursor.setPosition(QTextCursor.End, QTextCursor.MoveAnchor)
    editor.setTextCursor(cursor)

    editor.writer_docstring.write_docstring_for_shortcut()

    assert editor.toPlainText() == expected
